# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pneumonia Mnist dataset."""

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for Pneumonia Mnist dataset."""

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(shape=(28, 28, 1)),
            'label': tfds.features.ClassLabel(names=['Normal', 'Pneumonia']),
        }),
        supervised_keys=('image', 'label'),
        homepage='https://medmnist.com//',
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    npz_path = dl_manager.download(
        'https://zenodo.org/records/10519652/files/pneumoniamnist.npz'
    )

    with tf.io.gfile.GFile(npz_path, 'rb') as f:
      raw_data = np.load(f)

    train_images = np.expand_dims(raw_data.f.train_images, axis=-1)
    val_images = np.expand_dims(raw_data.f.val_images, axis=-1)
    test_images = np.expand_dims(raw_data.f.test_images, axis=-1)
    train_labels = raw_data.f.train_labels.flatten()
    val_labels = raw_data.f.val_labels.flatten()
    test_labels = raw_data.f.test_labels.flatten()

    return {
        'train': self._generate_examples(train_images, train_labels),
        'val': self._generate_examples(val_images, val_labels),
        'test': self._generate_examples(test_images, test_labels),
    }

  def _generate_examples(self, images, labels):
    """Yields examples."""
    for idx, (image, label) in enumerate(zip(images, labels)):
      yield idx, {
          'image': image,
          'label': int(np.squeeze(label)),
      }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BookSum summarization dataset."""

import json
import logging
import os
from typing import Dict, Iterator, Text, Tuple

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_DOCUMENT = "document"
_SUMMARY = "summary"
_SPLIT_FILENAMES = {
    "book": {
        "train": "book_summaries_aligned_train.jsonl",
        "validation": "book_summaries_aligned_val.jsonl",
        "test": "book_summaries_aligned_test.jsonl",
    },
    "chapter": {
        "train": "chapter_summary_aligned_train_split.jsonl",
        "validation": "chapter_summary_aligned_val_split.jsonl",
        "test": "chapter_summary_aligned_test_split.jsonl",
    },
}


class BooksumConfig(tfds.core.BuilderConfig):
  """BuilderConfig for BooksumConfig."""

  def __init__(self, *, granularity=None, **kwargs):
    """BuilderConfig for BooksumConfig.

    Args:
      granularity: str ("book", "chapter")
      **kwargs: keyword arguments forwarded to super.
    """
    super(BooksumConfig, self).__init__(**kwargs)
    self.granularity = granularity


class Builder(tfds.core.GeneratorBasedBuilder):
  """Booksum dataset builder."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }
  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  1) Go to https://github.com/salesforce/booksum, and run steps 1-3. Place the
     whole `booksum` git project in the manual folder.
  2) Download the chapterized books from https://storage.cloud.google.com/sfr-books-dataset-chapters-research/all_chapterized_books.zip
     and unzip to the manual folder.

  The manual folder should contain the following directories:

      - `booksum/`
      - `all_chapterized_books/`

  Note: Because the BookSum dataset is based on the availability of web-scraped
  data and may be incomplete, the `_generate_examples` method will automatically
  skip missing entries.
  """
  BUILDER_CONFIGS = [
      BooksumConfig(
          name="book",
          description="Book-level summarization",
          granularity="book",
      ),
      BooksumConfig(
          name="chapter",
          description="chapter-level summarization",
          granularity="chapter",
      ),
  ]

  def _info(self) -> tfds.core.DatasetInfo:
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            _DOCUMENT: tfds.features.Text(),
            _SUMMARY: tfds.features.Text(),
        }),
        supervised_keys=(_DOCUMENT, _SUMMARY),
        homepage="https://github.com/salesforce/booksum",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    granularity = self._builder_config.granularity  # pytype: disable=attribute-error  # always-use-return-annotations
    alignments_base_path = os.path.join(
        dl_manager.manual_dir,
        "booksum",
        "alignments",
        f"{granularity}-level-summary-alignments",
    )
    return {
        "train": self._generate_examples(
            alignments_path=os.path.join(
                alignments_base_path, _SPLIT_FILENAMES[granularity]["train"]
            ),
            base_path=dl_manager.manual_dir,
            granularity=granularity,
        ),
        "validation": self._generate_examples(
            alignments_path=os.path.join(
                alignments_base_path,
                _SPLIT_FILENAMES[granularity]["validation"],
            ),
            base_path=dl_manager.manual_dir,
            granularity=granularity,
        ),
        "test": self._generate_examples(
            alignments_path=os.path.join(
                alignments_base_path, _SPLIT_FILENAMES[granularity]["test"]
            ),
            base_path=dl_manager.manual_dir,
            granularity=granularity,
        ),
    }

  def _generate_examples(
      self,
      alignments_path: tfds.typing.PathLike,
      base_path: tfds.typing.PathLike,
      granularity: Text,
  ) -> Iterator[Tuple[Text, Dict[Text, Text]]]:
    """Yields examples."""
    with tf.io.gfile.GFile(alignments_path, "r") as f:
      for i, line in enumerate(f.read().strip().splitlines()):
        example_data = json.loads(line)
        input_path = os.path.join(
            base_path, example_data[f"{granularity}_path"]
        )
        summary_path = os.path.join(
            base_path, "booksum", "scripts", example_data["summary_path"]
        )
        if not tf.io.gfile.exists(input_path):
          logging.info("Skipping missing input: %s", input_path)
          continue
        if not tf.io.gfile.exists(summary_path):
          logging.info("Skipping missing summary: %s", summary_path)
          continue
        with tf.io.gfile.GFile(input_path, "r") as f:
          input_text = f.read().strip()
        with tf.io.gfile.GFile(summary_path, "r") as f:
          summary_text = " ".join(json.loads(f.read())["summary"]).strip()
        yield str(i), {
            _DOCUMENT: input_text,
            _SUMMARY: summary_text,
        }



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2024 Cohere and the current dataset script contributor.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Author Note: Data loader is heavily inspired by the builder in
# https://github.com/google-research/google-research/tree/main/lbpp_dataset
"""Cohere Less Basic Python Problems. All columns decoded."""

import base64
import json
import pickle
import zlib

from tensorflow_datasets.core.utils.lazy_imports_utils import pandas as pd
import tensorflow_datasets.public_api as tfds


_HOMEPAGE = "https://aclanthology.org/2024.findings-emnlp.772/"

_VERSION = tfds.core.Version("2.0.0")

_COLUMNS = [
    "task_id",
    "language",
    "title",
    "instruction",
    "completion",
    "test_file",
    "test_list",
    "signature",
    "categories",
    "test_setup",
]

_LANGUAGES = ["python", "cpp", "go", "java", "js", "rust"]
_ALL_LANGUAGE_ALIASES = ["all", "multilingual"]
_LANGUAGE_ALIAS_MAP = {
    "default": "python",
    "javascript": "js",
}


def decode_str(str_to_decode: str):
  return json.loads(
      pickle.loads(
          zlib.decompress(base64.b64decode(str_to_decode.encode("utf-8")))
      )
  )


class LBPPConfig(tfds.core.BuilderConfig):
  """BuilderConfig."""

  def __init__(self, name, description, features, **kwargs):
    super(LBPPConfig, self).__init__(name=name, version=_VERSION, **kwargs)
    self.name = name
    self.description = description
    self.features = features


class Builder(tfds.core.GeneratorBasedBuilder):
  """Builder for LBPP dataset."""

  VERSION = _VERSION
  LICENSE = "apache-2.0"
  BUILDER_CONFIGS = [
      LBPPConfig(
          name="all", description="Multilingual LBPP", features=_COLUMNS
      ),
      LBPPConfig(
          name="multilingual",
          description="Multilingual LBPP",
          features=_COLUMNS,
      ),
      LBPPConfig(name="default", description="Python LBPP", features=_COLUMNS),
      LBPPConfig(name="python", description="Python LBPP", features=_COLUMNS),
      LBPPConfig(name="cpp", description="C++ LBPP", features=_COLUMNS),
      LBPPConfig(name="go", description="Go LBPP", features=_COLUMNS),
      LBPPConfig(name="java", description="Java LBPP", features=_COLUMNS),
      LBPPConfig(name="js", description="JavaScript LBPP", features=_COLUMNS),
      LBPPConfig(
          name="javascript", description="JavaScript LBPP", features=_COLUMNS
      ),
      LBPPConfig(name="rust", description="JavaScript LBPP", features=_COLUMNS),
  ]
  DEFAULT_CONFIG_NAME = "python"

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "task_id": tfds.features.Text(),
            "language": tfds.features.Text(),
            "title": tfds.features.Text(),
            "instruction": tfds.features.Text(),
            "completion": tfds.features.Text(),
            "test_file": tfds.features.Text(),
            "test_list": tfds.features.Sequence(tfds.features.Text()),
            "signature": tfds.features.Text(),
            "categories": tfds.features.Sequence(tfds.features.Text()),
            "test_setup": tfds.features.Text(),
        }),
        homepage=_HOMEPAGE,
        supervised_keys=None,
    )

  def _split_generators(self, dl_manager):
    # Map alias to actual language
    data_loading_name = _LANGUAGE_ALIAS_MAP.get(
        self.builder_config.name, self.builder_config.name
    )
    hf_url_prefix = (
        "https://huggingface.co/datasets/CohereForAI/lbpp/resolve/main/"
    )
    if data_loading_name in _ALL_LANGUAGE_ALIASES:
      # Download all languages
      download_targets = [
          f"{hf_url_prefix}{lang}/test.parquet" for lang in _LANGUAGES
      ]
    else:
      download_targets = [f"{hf_url_prefix}{data_loading_name}/test.parquet"]

    downloaded_files = dl_manager.download(download_targets)

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={
                "filepaths": downloaded_files,
            },
        )
    ]

  def _generate_examples(self, filepaths: list[str]):
    key = 0
    for filepath in filepaths:
      df = pd.read_parquet(filepath)
      for line in df.to_dict(orient="records"):
        yield key, {
            "task_id": line["task_id"],
            "language": line["language"],
            "title": line["title"],
            "instruction": line["instruction"],
            "completion": decode_str(line["completion"]),
            "test_file": decode_str(line["test_file"]),
            "test_list": decode_str(line["test_list"]),
            "signature": line["signature"] or "",
            "categories": line["categories"],
            "test_setup": decode_str(line["test_setup"]),
        }
        key += 1



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""To write records into sharded records files."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, Sequence
import dataclasses
import functools
import itertools
import json
import os
import threading
from typing import Any, Callable

from etils import epy
from tensorflow_datasets.core.utils.lazy_imports_utils import apache_beam as beam

with epy.lazy_imports():
  # pylint: disable=g-import-not-at-top
  from absl import logging
  from etils import epath
  from tensorflow_datasets.core import example_parser
  from tensorflow_datasets.core import example_serializer
  from tensorflow_datasets.core import features as features_lib
  from tensorflow_datasets.core import file_adapters
  from tensorflow_datasets.core import hashing
  from tensorflow_datasets.core import naming
  from tensorflow_datasets.core import shuffle
  from tensorflow_datasets.core import utils
  from tensorflow_datasets.core.utils import file_utils
  from tensorflow_datasets.core.utils import shard_utils
  from tensorflow_datasets.core.utils import type_utils

  # pylint: enable=g-import-not-at-top

# TODO(tfds): Should be `TreeDict[FeatureValue]`
Example = Any
Key = int | bytes
KeyExample = tuple[Key, Example]

_INDEX_PATH_SUFFIX = "_index.json"


@dataclasses.dataclass(frozen=True)
class _ShardSpec:
  """Spec to write a final records shard.

  Attributes:
    shard_index: Index of the shard.
    path: The path where to write the shard.
    index_path: The path where to write index of the records in the
      corresponding shard. NOTE: Value for this attribute is always set, but
      usage depends on whether `write_examples` returned a list of record
      positions for each example.
    examples_number: Number of examples in shard.
    file_instructions: Reading instructions.
  """

  shard_index: int
  path: str
  index_path: str
  examples_number: int
  file_instructions: Sequence[shard_utils.FileInstruction]


def _raise_error_for_duplicated_keys(example1, example2, example_specs):
  """Log information about the examples and raise an AssertionError."""
  msg = "Two examples share the same hashed key!"
  logging.error(msg)
  try:
    parser = example_parser.ExampleParser(example_specs)
    ex1 = parser.parse_example(example1)
    ex2 = parser.parse_example(example2)
    logging.error("1st example: %s", ex1)
    logging.error("2nd example: %s", ex2)
  except ValueError:
    logging.error(
        "Failed to parse examples! Cannot log them to see the examples behind"
        " the duplicated keys. Raw example 1: %s, raw example 2: %s",
        example1,
        example2,
    )
  raise AssertionError(msg + " See logs above to view the examples.")


def _get_index_path(path: str) -> epath.PathLike:
  """Returns path to the index file of the records stored at the given path.

  E.g: Say the path to a shard of records (of a particular split) are stored at
  `your/path/to/records/foo.riegeli-00001-of-00005`, it is transformed into
  `your/path/to/records/foo.riegeli-00001-of-00005_index.json`.

  Args:
    path: Path to the record file.

  Returns:
    Path of the index file of a shard of records stored at given path.
  """
  return path + _INDEX_PATH_SUFFIX


def _get_shard_specs(
    num_examples: int,
    total_size: int,
    bucket_lengths: Sequence[int],
    filename_template: naming.ShardedFileTemplate,
    shard_config: shard_utils.ShardConfig,
) -> Sequence[_ShardSpec]:
  """Returns list of _ShardSpec instances, corresponding to shards to write.

  Args:
    num_examples: int, number of examples in split.
    total_size: int (bytes), sum of example sizes.
    bucket_lengths: list of ints, number of examples in each bucket.
    filename_template: template to format sharded filenames.
    shard_config: the configuration for creating shards.
  """
  num_shards = shard_config.get_number_shards(total_size, num_examples)
  shard_boundaries = shard_utils.get_shard_boundaries(num_examples, num_shards)
  shard_specs = []
  bucket_indexes = [str(i) for i in range(len(bucket_lengths))]
  from_ = 0
  for shard_index, to in enumerate(shard_boundaries):
    # Read the bucket indexes
    file_instructions = shard_utils.get_file_instructions(
        from_, to, bucket_indexes, bucket_lengths
    )
    shard_path = filename_template.sharded_filepath(
        shard_index=shard_index, num_shards=num_shards
    )
    index_path = _get_index_path(os.fspath(shard_path))
    shard_specs.append(
        _ShardSpec(
            shard_index=shard_index,
            path=os.fspath(shard_path),
            index_path=index_path,
            examples_number=to - from_,
            file_instructions=file_instructions,
        )
    )
    from_ = to
  return shard_specs


def _write_index_file(
    sharded_index_path: epath.PathLike, record_keys: Sequence[Any]
):
  """Writes index file for records of a shard at given `sharded_index_path`.

  NOTE: Each record position (i.e shard_key in shard_keys) is stored as a
  string for better readability of the index files. The reader should parse the
  position from string using `RecordPosition.from_str()` method.

  Args:
    sharded_index_path: Path to the sharded index path.
    record_keys: Sequence of keys/indices of the records in the shard.
  """
  # Store string representation of each record position.
  #
  # NOTE: This makes the index file more readable. Although the reader should
  # parse the record position from a string.
  index_info = {"index": [str(record_key) for record_key in record_keys]}
  epath.Path(sharded_index_path).write_text(json.dumps(index_info))
  logging.info("Wrote index file to %s", os.fspath(sharded_index_path))


class ExampleWriter:
  """Writes examples to files."""

  def __init__(self, file_format: file_adapters.FileFormat):
    self.file_format = file_format

  def write(
      self,
      path: epath.PathLike,
      examples: Iterable[type_utils.KeySerializedExample],
  ) -> file_adapters.ExamplePositions | None:
    """Write examples from iterator."""
    adapter = file_adapters.ADAPTER_FOR_FORMAT[self.file_format]
    return adapter.write_examples(path, examples)


class ThreadSafeIterator(Iterator):
  """A wrapper around a tee object to make it thread-safe.

  See https://stackoverflow.com/q/6703594 for more details.
  """

  def __init__(self, tee_object: Any, lock: threading.Lock):
    self._tee_object = tee_object
    self._lock = lock

  def __iter__(self):
    return self

  def __next__(self):
    with self._lock:
      return next(self._tee_object)

  def __copy__(self):
    return ThreadSafeIterator(self._tee_object.__copy__(), self._lock)


def thread_safe_tee(
    iterable: Iterable[Any], n: int
) -> tuple[ThreadSafeIterator, ...]:
  """Returns a tuple of n independent thread-safe iterators."""
  lock = threading.Lock()
  return tuple(
      ThreadSafeIterator(tee_object, lock)
      for tee_object in itertools.tee(iterable, n)
  )


class MultiOutputExampleWriter(ExampleWriter):
  """Example writer that can write multiple outputs."""

  def __init__(self, writers: Sequence[ExampleWriter]):  # pylint: disable=super-init-not-called
    self._writers: list[ExampleWriter] = list(writers)

  def add_writer(self, writer: ExampleWriter):
    self._writers.append(writer)

  def write(
      self,
      path: epath.PathLike,
      examples: Iterable[type_utils.KeySerializedExample],
  ) -> file_adapters.ExamplePositions | None:
    """Writes examples to multiple outputs."""
    write_fns = []
    for writer, my_iter in zip(
        self._writers, thread_safe_tee(examples, len(self._writers))
    ):
      if file_format := writer.file_format:
        shard_path = os.fspath(
            file_adapters.convert_path_to_file_format(
                path=path, file_format=file_format
            )
        )
        write_fns.append(functools.partial(writer.write, shard_path, my_iter))
      else:
        write_fns.append(functools.partial(writer.write, path, my_iter))

    for write_fn in write_fns:
      write_fn()


class ShardWriter:
  """Writes examples to a single shard."""

  def __init__(
      self,
      features: features_lib.FeatureConnector,
      serializer: example_serializer.Serializer,
      example_writer: ExampleWriter,
  ):
    """Initializes Writer.

    Args:
      features: the features of the dataset.
      serializer: class that can serialize examples.
      example_writer: class that writes examples to disk or elsewhere.
    """
    self._features = features
    self._serializer = serializer
    self._example_writer = example_writer

  def _serialize_example(self, example: Example) -> Any:
    """Encodes and serializes an example."""
    return self._serializer.serialize_example(
        self._features.encode_example(example)
    )

  def write(
      self,
      examples: Iterable[KeyExample],
      path: epath.Path,
  ) -> int:
    """Returns the number of examples written to the given path."""
    (for_writing, for_counting) = itertools.tee(examples, 2)

    def serialize_examples() -> Iterator[type_utils.KeySerializedExample]:
      for k, v in for_writing:
        yield k, self._serialize_example(v)

    self._example_writer.write(path=path, examples=serialize_examples())
    num_examples = sum(1 for _ in for_counting)

    return num_examples

  def write_with_beam(
      self,
      example_gen: Callable[[], Iterable[KeyExample]],
      path: epath.Path,
      shard_index: int,
      pipeline: beam.Pipeline,
  ) -> beam.Pipeline:
    """Writes a PCollection of examples to a file."""

    def write_examples(dummy_value: Any) -> tuple[int, int]:
      # The dummy value is needed to make the pipeline work with
      # `beam.Create([None])`.
      del dummy_value
      num_examples = self.write(examples=example_gen(), path=path)
      return shard_index, num_examples

    return (
        pipeline
        | f"CreateShard({path.name})" >> beam.Create([None])
        | f"WriteShard({path.name})" >> beam.Map(write_examples)
    )


class Writer:
  """Shuffles and writes Examples to sharded files.

  The number of shards is computed automatically.
  """

  def __init__(
      self,
      serializer: example_serializer.Serializer,
      filename_template: naming.ShardedFileTemplate,
      hash_salt,
      disable_shuffling: bool,
      example_writer: ExampleWriter,
      shard_config: shard_utils.ShardConfig | None = None,
      ignore_duplicates: bool = False,
  ):
    """Initializes Writer.

    Args:
      serializer: class that can serialize examples.
      filename_template: template to format sharded filenames.
      hash_salt (str or bytes): salt to hash keys.
      disable_shuffling (bool): Specifies whether to shuffle the records.
      example_writer: class that writes examples to disk or elsewhere.
      shard_config: the configuration for creating shards.
      ignore_duplicates: whether to ignore duplicated examples with the same
        key. If False, a `DuplicatedKeysError` will be raised on duplicates.
    """
    self._serializer = serializer
    self._shuffler = shuffle.Shuffler(
        dirpath=filename_template.data_dir,
        hash_salt=hash_salt,
        disable_shuffling=disable_shuffling,
        ignore_duplicates=ignore_duplicates,
    )
    self._filename_template = filename_template
    self._shard_config = shard_config or shard_utils.ShardConfig()
    self._example_writer = example_writer

  def write(self, key: int | bytes, example: Example):
    """Writes given example.

    The given example is not directly written to the shard, but to a
    temporary file (or memory). The finalize() method writes all the shards.

    Args:
      key (int|bytes): the key associated with the example. Used for shuffling.
      example: the Example to write to the shard.
    """
    serialized_example = self._serializer.serialize_example(example=example)
    self._shuffler.add(key, serialized_example)

  def finalize(self) -> tuple[list[int], int]:
    """Effectively writes examples to the shards."""
    if self._shuffler.num_examples == 0:
      raise AssertionError("No examples were yielded.")

    shard_specs = _get_shard_specs(
        num_examples=self._shuffler.num_examples,
        total_size=self._shuffler.size,
        bucket_lengths=self._shuffler.bucket_lengths,
        filename_template=self._filename_template,
        shard_config=self._shard_config,
    )
    filename = self._filename_template.sharded_filepaths_pattern()
    # Here we just loop over the examples, and don't use the instructions, just
    # the final number of examples in every shard. Instructions could be used to
    # parallelize, but one would need to be careful not to sort buckets twice.
    examples_generator = iter(
        utils.tqdm(
            self._shuffler,
            desc=f"Shuffling {filename}...",
            total=self._shuffler.num_examples,
            unit=" examples",
            leave=False,
            mininterval=1.0,
        )
    )
    try:
      for shard_spec in shard_specs:
        iterator = itertools.islice(
            examples_generator, 0, shard_spec.examples_number
        )
        record_keys = self._example_writer.write(shard_spec.path, iterator)

        # No shard keys returned (e.g: TFRecord format), index cannot be
        # created.
        if not record_keys:
          continue

        # Number of `shard_keys` received should match the number of examples
        # written in this shard.
        if len(record_keys) != int(shard_spec.examples_number):
          raise RuntimeError(
              f"Length of example `keys` ({len(record_keys)}) does not match "
              f"`shard_spec.examples_number: (`{shard_spec.examples_number})"
          )
        _write_index_file(shard_spec.index_path, record_keys)

    except shuffle.DuplicatedKeysError as err:
      _raise_error_for_duplicated_keys(
          err.item1, err.item2, self._serializer.example_specs
      )

    # Finalize the iterator to clear-up TQDM
    try:
      val = next(examples_generator)
    except StopIteration:
      pass
    else:
      raise ValueError(
          f"Shuffling more elements than expected. Additional element: {val}"
      )

    shard_lengths = [int(spec.examples_number) for spec in shard_specs]

    logging.info(
        "Done writing %s. Number of examples: %s (shards: %s)",
        filename,
        sum(shard_lengths),
        shard_lengths,
    )
    return shard_lengths, self._shuffler.size


@dataclasses.dataclass
class _ShardInfo:
  id: int
  num_examples: int
  size: int


class BeamWriter:
  """Shuffles / writes Examples beam collection to sharded files.

  Examples are not directly writen to the final shards, but first to temporary
  files. Only if that was successful, the shard is moved to the final location.
  """

  _OUTPUT_TAG_BUCKETS_LEN_SIZE = "tag_buckets_len_size"

  def __init__(
      self,
      serializer: example_serializer.Serializer,
      filename_template: naming.ShardedFileTemplate,
      hash_salt,
      disable_shuffling: bool,
      example_writer: ExampleWriter,
      shard_config: shard_utils.ShardConfig | None = None,
      ignore_duplicates: bool = False,
  ):
    """Init BeamWriter.

    Note that file "{filepath_prefix}.shard_lengths.json" is also created. It
    contains a list with the number of examples in each final shard. Eg:
    "[10,11,10,11]".

    Args:
      serializer: class that can serialize examples.
      filename_template: template to format sharded filenames.
      hash_salt: string, the salt to use for hashing of keys.
      disable_shuffling: bool, specifies whether to shuffle the records.
      example_writer: class that writes examples to storage.
      shard_config: the configuration for creating shards.
      ignore_duplicates: whether to ignore duplicated examples with the same
        key. If False, a `DuplicatedKeysError` will be raised on duplicates.
    """
    self._original_state = dict(
        serializer=serializer,
        filename_template=filename_template,
        hash_salt=hash_salt,
        disable_shuffling=disable_shuffling,
        shard_config=shard_config,
        example_writer=example_writer,
        ignore_duplicates=ignore_duplicates,
    )
    self._filename_template = filename_template
    self._split_info_path = (
        f"{filename_template.filepath_prefix()}.split_info.json"
    )
    self._serializer = serializer
    self._hasher = hashing.Hasher(hash_salt)
    self._split_info = None
    self._disable_shuffling = disable_shuffling
    self._shard_config = shard_config or shard_utils.ShardConfig()
    self._example_writer = example_writer
    self._ignore_duplicates = ignore_duplicates

  @functools.lru_cache()
  def _get_counter(self, name: str, namespace: str = "BeamWriter"):
    return beam.metrics.Metrics.counter(namespace, name)

  @functools.lru_cache()
  def _get_distribution(self, name: str, namespace: str = "BeamWriter"):
    return beam.metrics.Metrics.distribution(namespace, name)

  def inc_counter(self, name: str, value: int = 1) -> None:
    self._get_counter(name).inc(value)

  def __getstate__(self):
    return self._original_state

  def __setstate__(self, state):
    self.__init__(**state)

  def _serialize_example(
      self,
      key_example: tuple[hashing.HashKey, Example],
  ) -> tuple[Any, bytes]:
    """Returns (shard#, (hkey, serialized_example))."""
    key, example = key_example
    serialized_example = self._serializer.serialize_example(example)
    if self._disable_shuffling:
      hkey = key
    else:
      hkey = self._hasher.hash_key(key)
    self.inc_counter(name="serialized_examples")
    return (hkey, serialized_example)

  def _check_num_examples(self, num_examples: int) -> int:
    if num_examples <= 0:
      raise ValueError(
          f"The total number of generated examples is {num_examples} for split"
          f" {self._filename_template.split}. This should be >0!"
      )
    return num_examples

  def _write_final_shard(
      self,
      shardid_examples: tuple[int, Iterable[type_utils.KeySerializedExample]],
      non_empty_shard_ids: Sequence[int],
  ) -> _ShardInfo:
    """Write all examples of a shard to disk.

    Arguments:
      shardid_examples: tuple of the shard id and the serialized examples that
        belong to that shard.
      non_empty_shard_ids: list of the shard ids of all the non-empty shards.
        Must be sorted.

    Returns:
      the shard info of the written shard.
    """
    original_shard_id, examples = shardid_examples
    if not examples:
      raise AssertionError("Not a single example present in the PCollection!")
    # There may be empty shards, this ensures there are no gaps.
    shard_id = non_empty_shard_ids.index(original_shard_id)
    example_by_key = {}
    for key, example in examples:
      if key in example_by_key:
        if not self._ignore_duplicates:
          _raise_error_for_duplicated_keys(
              example_by_key[key], example, self._serializer.example_specs
          )
      else:
        example_by_key[key] = example
    shard_path = self._filename_template.sharded_filepath(
        shard_index=shard_id, num_shards=len(non_empty_shard_ids)
    )
    with utils.incomplete_files(epath.Path(shard_path)) as tmp_path:
      logging.info(
          "Writing %d examples to %s.", len(example_by_key), os.fspath(tmp_path)
      )
      record_keys = self._example_writer.write(
          tmp_path, sorted(example_by_key.items())
      )
    self.inc_counter(name="written_shards")
    # If there are record_keys, create index files.
    if record_keys:
      index_path = _get_index_path(os.fspath(shard_path))
      _write_index_file(index_path, list(record_keys))
    shard_size = sum(map(len, example_by_key.values()))
    return _ShardInfo(
        id=shard_id, num_examples=len(example_by_key), size=shard_size
    )

  def _number_of_shards(self, num_examples: int, total_size: int) -> int:
    """Returns the number of shards."""
    num_shards = self._shard_config.get_number_shards(
        total_size=total_size,
        num_examples=num_examples,
        uses_precise_sharding=False,
    )
    self.inc_counter(name="NumberOfShards", value=num_shards)
    return num_shards

  def _assign_shard(
      self,
      key_serialized_example: tuple[Any, bytes],
      num_shards: int,
      largest_key: Sequence[int],
  ) -> tuple[int, tuple[Any, bytes]]:
    """Assigns a shard id to the example."""
    key, _ = key_serialized_example
    largest_key = largest_key[0]
    shard_number = shuffle.get_bucket_number(
        hkey=key, num_buckets=num_shards, max_hkey=largest_key
    )
    self._get_distribution(name="ShardDistribution").update(shard_number)
    self.inc_counter(f"{self._filename_template.split}.shard_{shard_number}")
    return (shard_number, key_serialized_example)

  def _store_split_info(
      self,
      shard_infos: Sequence[_ShardInfo],
  ) -> None:
    """Stores the split info to disk."""
    shard_infos = sorted(shard_infos, key=lambda x: x.id)
    shard_lengths = [info.num_examples for info in shard_infos]
    total_size = sum([info.size for info in shard_infos])
    with utils.incomplete_file(epath.Path(self._split_info_path)) as tmp_path:
      tmp_path.write_text(
          json.dumps({"total_size": total_size, "shard_lengths": shard_lengths})
      )

  def _sort_shard_ids(
      self, shard_ids: Sequence[int], ideal_num_shards: int
  ) -> Sequence[int]:
    """Returns the sorted shard ids and logs information."""
    if len(shard_ids) != ideal_num_shards:
      logging.info(
          "Ideally there would be %d shards, but got %d non-empty shards.",
          ideal_num_shards,
          len(shard_ids),
      )
    return sorted(shard_ids)

  def write_from_pcollection(self, examples_pcollection):
    """Returns PTransform to write (key, example) PCollection."""
    serialized_examples = (
        examples_pcollection
        | "Serialize" >> beam.Map(self._serialize_example)
        # (key, serialized_example)
    )

    largest_key = (
        serialized_examples
        | beam.Keys()
        | "LargestKey" >> beam.combiners.Top.Largest(1)
    )
    num_examples = (
        serialized_examples
        | "CountExamples" >> beam.combiners.Count.Globally()
        | "CheckValidNumExamples" >> beam.Map(self._check_num_examples)
    )
    total_size = beam.pvalue.AsSingleton(
        serialized_examples
        | beam.Values()
        | beam.Map(len)
        | "TotalSize" >> beam.CombineGlobally(sum)
    )
    ideal_num_shards = beam.pvalue.AsSingleton(
        num_examples
        | "NumberOfShards"
        >> beam.Map(self._number_of_shards, total_size=total_size)
    )

    examples_per_shard = (
        serialized_examples
        | "AssignShard"
        >> beam.Map(
            self._assign_shard,
            largest_key=beam.pvalue.AsSingleton(largest_key),
            num_shards=ideal_num_shards,
        )
        # (shard_id, serialized_example)
        | "GroupShards" >> beam.GroupByKey()
        # (shard_id, [serialized_example])
    )

    # There may be shards that did not get assigned any examples. We will ignore
    # those shards and for that we need the ids of shards that are non-empty.
    non_empty_shard_ids = beam.pvalue.AsSingleton(
        examples_per_shard
        | "GetIdsOfNonEmptyShards" >> beam.Keys()
        | "CollectIdsOfNonEmptyShards" >> beam.transforms.combiners.ToList()
        | "SortIdsOfNonEmptyShards"
        >> beam.Map(self._sort_shard_ids, ideal_num_shards=ideal_num_shards)
    )

    return (
        examples_per_shard
        # (shard_id, [serialized_example])
        | "WriteFinalShards"
        >> beam.Map(
            self._write_final_shard, non_empty_shard_ids=non_empty_shard_ids
        )
        # (_ShardInfo)
        | "CollectShardInfo" >> beam.transforms.combiners.ToList()
        # [_ShardInfo]
        | "CalculateSplitInfo" >> beam.ParDo(self._store_split_info)
    )

  def finalize(self) -> tuple[list[int], int]:
    """Deletes tmp directory and returns shard_lengths and total_size.

    Returns:
      List of length <number of shards> containing the number of examples stored
      in each shard, and size of the files (in bytes).
    """
    if self._split_info is None:
      split_info_path = epath.Path(self._split_info_path)
      self._split_info = json.loads(split_info_path.read_bytes())
      split_info_path.unlink()

    return self._split_info["shard_lengths"], self._split_info["total_size"]


class NoShuffleBeamWriter:
  """Writes examples to sharded files using Beam in a non-deterministic way.

  The number of shards and in what shard an example is written is
  non-deterministic. This means that there may be a shards with few examples
  and other shards with many examples.

  This writer class should only be used when the ordering of the examples is not
  important, e.g., when a file format that supports random access is used.

  The speed of writing is faster than the Writer class because it does not need
  to shuffle the examples and make sure that the examples are written to the
  correct shard.
  """

  _OUTPUT_TAG_BUCKETS_LEN_SIZE = "tag_buckets_len_size"

  def __init__(
      self,
      serializer: example_serializer.Serializer,
      filename_template: naming.ShardedFileTemplate,
      file_format: file_adapters.FileFormat,
      num_shards: int | None = None,
  ):
    """Init BeamWriter.

    Note that file "{filepath_prefix}.shard_lengths.json" is also created. It
    contains a list with the number of examples in each final shard. Eg:
    "[10,11,10,11]".

    Args:
      serializer: class that can serialize examples.
      filename_template: template to format sharded filenames.
      file_format: the file format to use.
      num_shards: the number of shards to use. If `None`, then the number of
        shards is calculated automatically.
    """
    self._original_state = dict(
        serializer=serializer,
        filename_template=filename_template,
        file_format=file_format,
    )
    self._file_format = file_format
    self._file_adapter = file_adapters.ADAPTER_FOR_FORMAT[self._file_format]
    self._filename_template = filename_template
    self._serializer = serializer
    self._num_shards = num_shards

  @functools.lru_cache()
  def _get_counter(self, name: str, namespace: str = "BeamWriter"):
    return beam.metrics.Metrics.counter(namespace, name)

  def inc_counter(self, name: str, value: int = 1) -> None:
    self._get_counter(name).inc(value)

  def __getstate__(self):
    return self._original_state

  def __setstate__(self, state):
    self.__init__(**state)

  def _serialize_example(
      self,
      key_example: tuple[hashing.HashKey, Example],
  ) -> bytes:
    """Returns (serialized_example)."""
    _, example = key_example
    self.inc_counter(name="serialized_examples")
    return self._serializer.serialize_example(example)

  def write_from_pcollection(self, examples_pcollection):
    """Returns PTransform to write (key, example) PCollection."""
    serialized_examples = (
        examples_pcollection
        | "Shuffle" >> beam.Reshuffle()
        | "Serialize" >> beam.Map(self._serialize_example)
    )
    return serialized_examples | "Write" >> self._file_adapter.beam_sink(
        filename_template=self._filename_template, num_shards=self._num_shards
    )

  def finalize(self) -> tuple[list[int], int]:
    """Returns the computed shard_lengths and total_size.

    Returns:
      List of length <number of shards> containing the number of examples stored
      in each shard, and size of the files (in bytes).
    """
    logging.info("Finalizing writer for %s", self._filename_template.split)
    shard_lengths_and_sizes = self._file_adapter.shard_lengths_and_sizes(
        self._filename_template, num_shards=self._num_shards
    )
    shard_lengths = [length for length, _ in shard_lengths_and_sizes]
    total_size_bytes = sum(size for _, size in shard_lengths_and_sizes)

    logging.info(
        "Found %d shards with a total size of %d bytes.",
        len(shard_lengths),
        total_size_bytes,
    )

    # Empty shards may be produced by Beam. We delete them and rename the
    # non-empty shards accordingly.
    all_shard_paths = self._filename_template.sharded_filepaths(
        len(shard_lengths)
    )
    non_empty_shards: list[epath.Path] = []
    non_empty_shard_lengths: list[int] = []
    empty_shards: list[epath.Path] = []
    for length, shard_path in zip(shard_lengths, all_shard_paths):
      if length > 0:
        non_empty_shard_lengths.append(length)
        non_empty_shards.append(shard_path)
      else:
        empty_shards.append(shard_path)

    non_empty_shard_paths = self._filename_template.sharded_filepaths(
        len(non_empty_shards)
    )
    if empty_shards:
      old_paths: list[epath.Path] = []
      new_paths: list[epath.Path] = []
      for orig_path, new_path in zip(non_empty_shards, non_empty_shard_paths):
        old_paths.append(orig_path)
        new_paths.append(new_path)
      file_utils.bulk_delete(empty_shards)
      file_utils.bulk_rename(old_paths=old_paths, new_paths=new_paths)

    return non_empty_shard_lengths, total_size_bytes



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Imagenet datasets."""

import io
import os

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.datasets.imagenet2012 import imagenet_common
import tensorflow_datasets.public_api as tfds

# Web-site is asking to cite paper from 2015.
# https://image-net.org/challenges/LSVRC/2012/index#cite

# From https://github.com/cytsai/ilsvrc-cmyk-image-list
CMYK_IMAGES = [
    'n01739381_1309.JPEG',
    'n02077923_14822.JPEG',
    'n02447366_23489.JPEG',
    'n02492035_15739.JPEG',
    'n02747177_10752.JPEG',
    'n03018349_4028.JPEG',
    'n03062245_4620.JPEG',
    'n03347037_9675.JPEG',
    'n03467068_12171.JPEG',
    'n03529860_11437.JPEG',
    'n03544143_17228.JPEG',
    'n03633091_5218.JPEG',
    'n03710637_5125.JPEG',
    'n03961711_5286.JPEG',
    'n04033995_2932.JPEG',
    'n04258138_17003.JPEG',
    'n04264628_27969.JPEG',
    'n04336792_7448.JPEG',
    'n04371774_5854.JPEG',
    'n04596742_4225.JPEG',
    'n07583066_647.JPEG',
    'n13037406_4650.JPEG',
]

PNG_IMAGES = ['n02105855_2933.JPEG']


class Builder(tfds.core.GeneratorBasedBuilder):
  """Imagenet 2012, aka ILSVRC 2012."""

  VERSION = tfds.core.Version('5.1.0')
  SUPPORTED_VERSIONS = [
      tfds.core.Version('5.0.0'),
  ]
  RELEASE_NOTES = {
      '5.1.0': 'Added test split.',
      '5.0.0': 'New split API (https://tensorflow.org/datasets/splits)',
      '4.0.0': '(unpublished)',
      '3.0.0': """
      Fix colorization on ~12 images (CMYK -> RGB).
      Fix format for consistency (convert the single png image to Jpeg).
      Faster generation reading directly from the archive.
      """,
      '2.0.1': 'Encoding fix. No changes from user point of view.',
      '2.0.0': 'Fix validation labels.',
  }

  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  manual_dir should contain two files: ILSVRC2012_img_train.tar and
  ILSVRC2012_img_val.tar.
  You need to register on https://image-net.org/download-images in order
  to get the link to download the dataset.
  """

  def _info(self):
    names_file = imagenet_common.label_names_file()
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'image': tfds.features.Image(encoding_format='jpeg'),
            'label': tfds.features.ClassLabel(names_file=names_file),
            'file_name': tfds.features.Text(),  # Eg: 'n15075141_54.JPEG'
        }),
        supervised_keys=('image', 'label'),
        homepage='https://image-net.org/',
    )

  def _split_generators(self, dl_manager):
    train_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_train.tar')
    val_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_val.tar')
    test_path = os.path.join(dl_manager.manual_dir, 'ILSVRC2012_img_test.tar')
    splits = []
    _add_split_if_exists(
        split_list=splits,
        split=tfds.Split.TRAIN,
        split_path=train_path,
        dl_manager=dl_manager,
    )
    _add_split_if_exists(
        split_list=splits,
        split=tfds.Split.VALIDATION,
        split_path=val_path,
        dl_manager=dl_manager,
        validation_labels=imagenet_common.get_validation_labels(val_path),
    )
    _add_split_if_exists(
        split_list=splits,
        split=tfds.Split.TEST,
        split_path=test_path,
        dl_manager=dl_manager,
        labels_exist=False,
    )
    if not splits:
      raise AssertionError(
          'ImageNet requires manual download of the data. Please download '
          'the data and place them into:\n'
          f' * train: {train_path}\n'
          f' * test: {test_path}\n'
          f' * validation: {val_path}\n'
          'At least one of the split should be available.'
      )
    return splits

  def _fix_image(self, image_fname, image):
    """Fix image color system and format starting from v 3.0.0."""
    if self.version < '3.0.0':
      return image
    if image_fname in CMYK_IMAGES:
      image = io.BytesIO(tfds.core.utils.jpeg_cmyk_to_rgb(image.read()))
    elif image_fname in PNG_IMAGES:
      image = io.BytesIO(tfds.core.utils.png_to_jpeg(image.read()))
    return image

  def _generate_examples(
      self, archive, validation_labels=None, labels_exist=True
  ):
    """Yields examples."""
    if not labels_exist:  # Test split
      for key, example in imagenet_common.generate_examples_test(archive):
        yield key, example
    if validation_labels:  # Validation split
      for key, example in imagenet_common.generate_examples_validation(
          archive, validation_labels
      ):
        yield key, example
    # Training split. Main archive contains archives names after a synset noun.
    # Each sub-archive contains pictures associated to that synset.
    for fname, fobj in archive:
      label = fname[:-4]  # fname is something like 'n01632458.tar'
      # TODO(b/117643231): in py3, the following lines trigger tarfile module
      # to call `fobj.seekable()`, which Gfile doesn't have. We should find an
      # alternative, as this loads ~150MB in RAM.
      fobj_mem = io.BytesIO(fobj.read())
      for image_fname, image in tfds.download.iter_archive(
          fobj_mem, tfds.download.ExtractMethod.TAR_STREAM
      ):
        image = self._fix_image(image_fname, image)
        record = {
            'file_name': image_fname,
            'image': image,
            'label': label,
        }
        yield image_fname, record


def _add_split_if_exists(split_list, split, split_path, dl_manager, **kwargs):
  """Add split to given list of splits only if the file exists."""
  if not tf.io.gfile.exists(split_path):
    logging.warning(
        (
            'ImageNet 2012 Challenge %s split not found at %s. '
            'Proceeding with data generation anyways but the split will be '
            'missing from the dataset...'
        ),
        str(split),
        split_path,
    )
  else:
    split_list.append(
        tfds.core.SplitGenerator(
            name=split,
            gen_kwargs={
                'archive': dl_manager.iter_archive(split_path),
                **kwargs,
            },
        ),
    )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Datasets of the Open Images Challange 2019.

https://storage.googleapis.com/openimages/web/challenge2019.html
"""

from __future__ import annotations

import abc

import numpy as np
import tensorflow_datasets.public_api as tfds

_URL = "https://storage.googleapis.com/openimages/web/challenge2019.html"

_GOOGLE_URL_PREFIX = (
    "https://storage.googleapis.com/openimages/challenge_2019/challenge-2019-"
)
_FIGURE_EIGHT_BASE_URL = (
    "https://datasets.figure-eight.com/figure_eight_datasets/open-images/"
)
_TRAIN_IMAGES_URLS = [
    "{}zip_files_copy/train_{:02d}.zip".format(_FIGURE_EIGHT_BASE_URL, n)
    for n in range(9)
]
_VALIDATION_IMAGES_URL = (
    _FIGURE_EIGHT_BASE_URL + "zip_files_copy/validation.zip"
)
_TEST_IMAGES_URL = _FIGURE_EIGHT_BASE_URL + "test_challenge.zip"
_NUM_CLASSES = 500


class OpenImagesChallenge2019Config(tfds.core.BuilderConfig):
  """BuilderConfig for OpenImages Challenge 2019 datasets."""

  def __init__(self, target_pixels=None, **kwargs):
    kwargs.setdefault("version", tfds.core.Version("1.0.0"))
    super(OpenImagesChallenge2019Config, self).__init__(**kwargs)
    self._target_pixels = target_pixels

  @property
  def target_pixels(self):
    return self._target_pixels


class _OpenImagesChallenge2019(tfds.core.BeamBasedBuilder):  # pytype: disable=ignored-abstractmethod  # abcmeta-check
  """Base abstract class for Open Images Challenge 2019 datasets."""

  BUILDER_CONFIGS = [
      OpenImagesChallenge2019Config(
          name="200k",
          description="Images have at most 200,000 pixels, at 72 JPEG quality.",
          target_pixels=200000,
      ),
      OpenImagesChallenge2019Config(
          name="300k",
          description="Images have at most 300,000 pixels, at 72 JPEG quality.",
          target_pixels=300000,
      ),
  ]

  @property
  @abc.abstractmethod
  def annotation_urls(self):
    """Dictionary passed to the DownloadManager to download annotations.

    An example:
      {"test_annotations": "https://somewebpage.com/data/openimages/test.txt"}

    Returns:
      A dictionary whose values are the URLs to download the annotations of the
      dataset, and the keys are some short string identifying the URL.
      This dictionary is passed to the DownloadManager.
    """

  def _split_generators(self, dl_manager):
    urls = {
        "train_images": _TRAIN_IMAGES_URLS,
        "test_images": [_TEST_IMAGES_URL],
        "validation_images": [_VALIDATION_IMAGES_URL],
    }
    urls.update(self.annotation_urls)  # pytype: disable=wrong-arg-types
    paths = dl_manager.download(urls)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(paths=paths, split="train"),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs=dict(paths=paths, split="test"),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs=dict(paths=paths, split="validation"),
        ),
    ]


class Builder(_OpenImagesChallenge2019):
  """Dataset for the Detection Track."""

  @property
  def annotation_urls(self):
    return {
        "train_image_label": (
            _GOOGLE_URL_PREFIX + "train-detection-human-imagelabels.csv"
        ),
        "train_boxes": _GOOGLE_URL_PREFIX + "train-detection-bbox.csv",
        "validation_image_label": (
            _GOOGLE_URL_PREFIX + "validation-detection-human-imagelabels.csv"
        ),
        "validation_boxes": (
            _GOOGLE_URL_PREFIX + "validation-detection-bbox.csv"
        ),
        "classes": _GOOGLE_URL_PREFIX + "classes-description-500.csv",
        "hierarchy": _GOOGLE_URL_PREFIX + "label500-hierarchy.json",
    }

  def _info(self):
    label = tfds.features.ClassLabel(num_classes=_NUM_CLASSES)
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "id": tfds.features.Text(),
            "image": tfds.features.Image(),
            # A sequence of image-level labels.
            "objects": tfds.features.Sequence({
                "label": label,
                # All labels have been verified by humans.
                #  - If confidence is 1.0, the object IS in the image.
                #  - If confidence is 0.0, the object is NOT in the image.
                "confidence": np.float32,
                "source": tfds.features.Text(),
            }),
            # A sequence of bounding boxes.
            "bobjects": tfds.features.Sequence({
                "label": label,
                "bbox": tfds.features.BBoxFeature(),
                "is_group_of": np.bool_,
            }),
        }),
        homepage=_URL,
    )

  def _build_pcollection(self, pipeline, paths, split):
    beam = tfds.core.lazy_imports.apache_beam
    # We need to lazily import the oi_beam module (and thus, violate the
    # "imports only at the top" rule), so that beam is only required during the
    # generation of the dataset, and not to use the dataset itself (once built).
    # See: https://www.tensorflow.org/datasets/beam_datasets.
    import tensorflow_datasets.object_detection.open_images_challenge2019_beam as oi_beam  # pylint: disable=g-import-not-at-top,import-outside-toplevel

    if split == "test":
      # Note: annotations are not available for the test split.
      generate_examples_kwargs = dict(
          image_labels_filepath=None,
          box_labels_filepath=None,
          hierarchy_filepath=None,
          classes_filepath=None,
      )
    else:
      generate_examples_kwargs = dict(
          image_labels_filepath=paths["{}_image_label".format(split)],
          box_labels_filepath=paths["{}_boxes".format(split)],
          hierarchy_filepath=paths["hierarchy"],
          classes_filepath=paths["classes"],
      )
    # Fill class names after the data has been downloaded.
    oi_beam.fill_class_names_in_tfds_info(paths["classes"], self.info.features)
    return (
        pipeline
        | beam.Create(paths["{}_images".format(split)])
        | "ReadImages" >> beam.ParDo(oi_beam.ReadZipFn())
        | "ProcessImages"
        >> beam.ParDo(
            oi_beam.ProcessImageFn(
                target_pixels=self.builder_config.target_pixels, jpeg_quality=72
            )
        )
        | "GenerateExamples"
        >> beam.ParDo(
            oi_beam.CreateDetectionExampleFn(**generate_examples_kwargs)
        )
    )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""robomimic_ph dataset."""

from tensorflow_datasets.datasets.robomimic_mh import robomimic_mh_dataset_builder
import tensorflow_datasets.public_api as tfds


class RobomimicMhTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for robomimic_mh dataset."""

  DATASET_CLASS = robomimic_mh_dataset_builder.Builder
  SPLITS = {'train': 1}
  SKIP_TF1_GRAPH_MODE = True

  DL_EXTRACT_RESULT = {'file_path': 'lift_image.hdf5'}
  DL_DOWNLOAD_RESULT = {'file_path': 'mh/lift_image.hdf5'}

  BUILDER_CONFIG_NAMES_TO_TEST = ['lift_mh_image']



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parquet DataSource class."""

from collections.abc import Iterable, Sequence
import dataclasses
from typing import Any

import numpy as np
from tensorflow_datasets.core.data_sources import base
from tensorflow_datasets.core.utils.lazy_imports_utils import parquet as pq
from tensorflow_datasets.core.utils.lazy_imports_utils import pyarrow as pa


@dataclasses.dataclass
class _ParquetTable:
  """Overloads pa.Table to transform it into a valid data source."""

  table: pa.Table

  def __getitems__(self, keys: Iterable[int]) -> Sequence[Any]:
    """Retrieves ranges/slices/lists of indexes from the Parquet table."""
    keys = np.fromiter(keys, np.int64)
    if not keys.size:
      return []
    # All elements are written in the first column (see core.ParquetFileAdapter)
    elements = self.table.take(keys).column(0)
    return [memoryview(element.as_buffer()) for element in elements]

  def __getitem__(self, key: int) -> Any:
    """Retrieves the n-th element from the Parquet table."""
    # The element is written in the first column (see core.ParquetFileAdapter)
    element = self.table.slice(key, 1).column(0)
    if len(element) == 1:
      return memoryview(element[0].as_buffer())
    raise IndexError(f'Could not find element at index {key}')

  def __len__(self) -> int:
    return self.table.num_rows


@dataclasses.dataclass(repr=False)
class ParquetDataSource(base.BaseDataSource):
  """ParquetDataSource to read from a ParquetDataset."""

  def __post_init__(self):
    file_instructions = base.file_instructions(self.dataset_info, self.split)
    filenames = [
        file_instruction.filename for file_instruction in file_instructions
    ]
    dataset = pq.ParquetDataset(filenames)
    table = dataset.read()
    self.data_source = _ParquetTable(table=table)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for universal_dependencies_utils."""

from tensorflow_datasets.datasets.universal_dependencies import universal_dependencies_utils as ud_utils


def test_prepare_ud_filepaths():
  expected_output = ["https://a/b"]

  filepaths_1 = ud_utils.prepare_ud_filepaths(
      path_prefix="https://a", filepaths="b"
  )
  filepaths_2 = ud_utils.prepare_ud_filepaths(
      path_prefix="https://a", filepaths=["b"]
  )

  assert filepaths_1 == expected_output
  assert filepaths_2 == expected_output



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from etils import epath
from tensorflow_datasets.core import dataset_info as dataset_info_lib
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.scripts.cli import convert_format_utils


def _create_dataset_info(
    data_dir: epath.Path,
    name: str = 'a',
    version: str = '1.2.3',
    config_name: str = 'cfg1',
    file_format: file_adapters.FileFormat = file_adapters.FileFormat.TFRECORD,
    split_lengths: dict[str, int] | None = None,
) -> dataset_info_pb2.DatasetInfo:
  filename_template = naming.ShardedFileTemplate(
      data_dir=data_dir,
      dataset_name=name,
      split=None,  # will be set in _create_split_info
      filetype_suffix=file_format.value,
  )
  split_lengths = split_lengths or {}
  info = dataset_info_lib.DatasetInfo(
      builder=dataset_info_lib.DatasetIdentity(
          name=name,
          version=version,
          config_name=config_name,
          data_dir=data_dir,
          module_name='xyz',
      ),
      split_dict=splits_lib.SplitDict(
          split_infos=[
              _create_split_info(
                  name=split_name,
                  shard_lengths=[1] * split_length,
                  filename_template=filename_template,
              )
              for split_name, split_length in split_lengths.items()
          ]
      ),
  )
  info.set_file_format(file_format)
  data_dir.mkdir(parents=True, exist_ok=True)
  info.write_to_directory(data_dir)
  return info.as_proto


def _create_split_info(
    name: str,
    shard_lengths: list[int],
    filename_template: naming.ShardedFileTemplate,
) -> splits_lib.SplitInfo:
  return splits_lib.SplitInfo(
      name=name,
      shard_lengths=shard_lengths,
      num_bytes=1,
      filename_template=filename_template.replace(split=name),
  )


def test_shard_instructions_for_split():
  split_info = splits_lib.SplitInfo(
      name='train',
      shard_lengths=[1, 2, 3],
      num_bytes=1,
      filename_template=naming.ShardedFileTemplate(
          data_dir=epath.Path('/in'),
          dataset_name='ds',
          split='train',
          filetype_suffix='tfrecord',
      ),
  )
  convert_config = convert_format_utils.ConvertConfig(
      in_file_format=file_adapters.FileFormat.TFRECORD,
      out_file_format=file_adapters.FileFormat.ARRAY_RECORD,
  )
  actual = convert_format_utils._shard_instructions_for_split(
      split_info=split_info,
      out_path=epath.Path('/out'),
      convert_config=convert_config,
  )
  assert actual == [
      convert_format_utils.ShardInstruction(
          in_path=epath.Path('/in/ds-train.tfrecord-00000-of-00003'),
          out_path=epath.Path('/out/ds-train.array_record-00000-of-00003'),
          config=convert_config,
      ),
      convert_format_utils.ShardInstruction(
          in_path=epath.Path('/in/ds-train.tfrecord-00001-of-00003'),
          out_path=epath.Path('/out/ds-train.array_record-00001-of-00003'),
          config=convert_config,
      ),
      convert_format_utils.ShardInstruction(
          in_path=epath.Path('/in/ds-train.tfrecord-00002-of-00003'),
          out_path=epath.Path('/out/ds-train.array_record-00002-of-00003'),
          config=convert_config,
      ),
  ]




def test_create_out_dir():
  actual = convert_format_utils._create_out_dir(
      dataset_dir='/a/b/c/d',
      root_in_dir='/a/b',
      root_out_dir='/e',
  )
  assert os.fspath(actual) == '/e/c/d'


def test_create_from_to_dirs():
  references = [
      naming.DatasetReference(
          dataset_name='a', config='cfg1', version='1.0.0', data_dir='/data/in'
      ),
      naming.DatasetReference(
          dataset_name='a', config='cfg2', version='1.0.0', data_dir='/data/in'
      ),
      naming.DatasetReference(
          dataset_name='b', config='cfg1', version='1.0.0', data_dir='/data/in'
      ),
      naming.DatasetReference(
          dataset_name='c', version='1.0.0', data_dir='/data/in'
      ),
  ]
  actual = convert_format_utils._create_from_to_dirs(
      references, epath.Path('/data/in'), epath.Path('/out')
  )
  assert actual == {
      epath.Path('/data/in/a/cfg1/1.0.0'): epath.Path('/out/a/cfg1/1.0.0'),
      epath.Path('/data/in/a/cfg2/1.0.0'): epath.Path('/out/a/cfg2/1.0.0'),
      epath.Path('/data/in/b/cfg1/1.0.0'): epath.Path('/out/b/cfg1/1.0.0'),
      epath.Path('/data/in/c/1.0.0'): epath.Path('/out/c/1.0.0'),
  }


def test_create_from_to_dirs_same_folder():
  references = [
      naming.DatasetReference(
          dataset_name='a', config='cfg1', version='1.0.0', data_dir='/data/in'
      )
  ]
  actual = convert_format_utils._create_from_to_dirs(
      references, root_in_dir=epath.Path('/data/in'), out_path=None
  )
  expected_path = epath.Path('/data/in/a/cfg1/1.0.0')
  assert actual == {expected_path: expected_path}


def test_get_root_data_dir(tmpdir):
  leaf_data_dir = epath.Path(tmpdir) / 'a/cfg1/1.2.3'
  info = _create_dataset_info(data_dir=leaf_data_dir)
  actual = convert_format_utils._get_root_data_dir(
      in_dir=leaf_data_dir, info=info
  )
  assert os.fspath(tmpdir) == os.fspath(actual)


def test_record_source_dataset(tmpdir):
  tmpdir = epath.Path(tmpdir)
  in_root_data_dir = tmpdir / 'data'
  in_data_dir = in_root_data_dir / 'a/cfg1/1.2.3'
  out_data_dir = tmpdir / 'out'
  in_data_dir.mkdir(parents=True)
  out_data_dir.mkdir(parents=True)
  info = _create_dataset_info(in_data_dir)
  convert_config = convert_format_utils.ConvertConfig(
      out_file_format=file_adapters.FileFormat.RIEGELI
  )
  convert_format_utils.convert_metadata(
      in_dir=in_data_dir,
      out_path=out_data_dir,
      info=info,
      convert_config=convert_config,
  )
  converted_info = dataset_info_lib.read_proto_from_builder_dir(out_data_dir)
  assert converted_info.name == info.name
  assert converted_info.version == info.version
  assert converted_info.config_name == info.config_name
  assert converted_info.module_name == info.module_name
  assert converted_info.file_format == file_adapters.FileFormat.RIEGELI.value
  assert len(converted_info.data_source_accesses) == 1
  data_source_access = converted_info.data_source_accesses[0]
  assert data_source_access.tfds_dataset.data_dir == os.fspath(in_root_data_dir)
  assert data_source_access.tfds_dataset.name == info.name
  assert data_source_access.tfds_dataset.version == info.version
  assert data_source_access.tfds_dataset.config == info.config_name


def test_convert_metadata_add_to_existing(tmpdir):
  in_data_dir = epath.Path(tmpdir) / 'a/cfg1/1.2.3'
  in_data_dir.mkdir(parents=True)
  info = _create_dataset_info(data_dir=in_data_dir, split_lengths={'train': 1})

  # Create a converted shard in the input directory.
  converted_shard = in_data_dir / 'a-train.riegeli-00000-of-00001'
  converted_shard.touch()
  convert_config = convert_format_utils.ConvertConfig(
      out_file_format=file_adapters.FileFormat.RIEGELI
  )

  convert_format_utils.convert_metadata(
      in_dir=in_data_dir,
      out_path=in_data_dir,
      info=info,
      convert_config=convert_config,
  )
  converted_info = dataset_info_lib.read_proto_from_builder_dir(in_data_dir)
  assert converted_info.file_format == file_adapters.FileFormat.TFRECORD.value
  assert converted_info.alternative_file_formats == [
      file_adapters.FileFormat.RIEGELI.value
  ]


def test_convert_metadata_missing_shards(tmpdir):
  in_data_dir = epath.Path(tmpdir) / 'a/cfg1/1.2.3'
  in_data_dir.mkdir(parents=True)
  info = _create_dataset_info(
      in_data_dir, split_lengths={'train': 2, 'test': 1}
  )
  convert_config = convert_format_utils.ConvertConfig(
      out_file_format=file_adapters.FileFormat.RIEGELI
  )
  convert_format_utils.convert_metadata(
      in_dir=in_data_dir,
      out_path=in_data_dir,
      info=info,
      convert_config=convert_config,
  )
  converted_info = dataset_info_lib.read_proto_from_builder_dir(in_data_dir)
  assert converted_info.file_format == file_adapters.FileFormat.TFRECORD.value
  # Make sure no alternative file format was added.
  assert not converted_info.alternative_file_formats



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GTZAN Music Speech dataset."""

from tensorflow_datasets import testing
from tensorflow_datasets.audio.gtzan_music_speech import gtzan_music_speech


class GTZANTest(testing.DatasetBuilderTestCase):
  DATASET_CLASS = gtzan_music_speech.GTZANMusicSpeech
  SPLITS = {
      "train": 1,  # Number of fake train examples
  }
  DL_EXTRACT_RESULT = {"music_speech": ""}


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Wrapper around `tfds build`."""

from absl import app
from absl import flags
from tensorflow_datasets.scripts.cli import main as main_cli


def _parse_flags(argv: list[str]) -> main_cli.Args:
  """Command lines flag parsing."""
  return main_cli._parse_flags([argv[0], 'build'] + argv[1:])  # pylint: disable=protected-access


def main(args: main_cli.Args) -> None:
  from absl import logging
  logging.warning(
      '***`tfds build` should be used instead of `download_and_prepare`.***'
  )
  main_cli.main(args)


if __name__ == '__main__':
  app.run(main, flags_parser=_parse_flags)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Celeba-HQ dataset."""

import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds


class CelebaHQConfig(tfds.core.BuilderConfig):
  """BuilderConfig for CelebaHQ."""

  def __init__(self, *, resolution, **kwargs):
    """BuilderConfig for SQUAD.

    Args:
      resolution: Resolution of the image. Values supported: powers of 2 up to
        1024.
      **kwargs: keyword arguments forwarded to super.
    """
    v2 = tfds.core.Version("2.0.0")
    super(CelebaHQConfig, self).__init__(
        name="%d" % resolution,
        description="CelebaHQ images in %d x %d resolution"
        % (resolution, resolution),
        version=v2,
        release_notes={
            "2.0.0": "New split API (https://tensorflow.org/datasets/splits)",
        },
        **kwargs,
    )
    self.resolution = resolution
    self.file_name = "data%dx%d.tar" % (resolution, resolution)


class Builder(tfds.core.GeneratorBasedBuilder):
  """Celeba_HQ Dataset."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  manual_dir should contain multiple tar files with images (data2x2.tar,
  data4x4.tar .. data1024x1024.tar).
  Detailed instructions are here:
  https://github.com/tkarras/progressive_growing_of_gans#preparing-datasets-for-training
  """

  BUILDER_CONFIGS = [
      CelebaHQConfig(resolution=1024),
      CelebaHQConfig(resolution=512),
      CelebaHQConfig(resolution=256),
      CelebaHQConfig(resolution=128),
      CelebaHQConfig(resolution=64),
      CelebaHQConfig(resolution=32),
      CelebaHQConfig(resolution=16),
      CelebaHQConfig(resolution=8),
      CelebaHQConfig(resolution=4),
      CelebaHQConfig(resolution=2),
      CelebaHQConfig(resolution=1),
  ]

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {
                "image": tfds.features.Image(
                    shape=(
                        self.builder_config.resolution,
                        self.builder_config.resolution,
                        3,
                    ),
                    encoding_format="png",
                ),
                "image/filename": tfds.features.Text(),
            },
        ),
        homepage="https://github.com/tkarras/progressive_growing_of_gans",
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    image_tar_file = os.path.join(
        dl_manager.manual_dir, self.builder_config.file_name
    )
    if not tf.io.gfile.exists(image_tar_file):
      # The current celebahq generation code depends on a concrete version of
      # pillow library and cannot be easily ported into tfds.
      msg = "You must download the dataset files manually and place them in: "
      msg += dl_manager.manual_dir
      msg += " as .tar files. See testing/test_data/fake_examples/celeb_a_hq "
      raise AssertionError(msg)
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={"archive": dl_manager.iter_archive(image_tar_file)},
        )
    ]

  def _generate_examples(self, archive):
    for fname, fobj in archive:
      record = {"image": fobj, "image/filename": fname}
      yield fname, record



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for interacting with BoundingBox features.

This script contains various functions to convert commonly used bounding boxes
formats to the YXYX format, which is used by TFDS.

A list of such formats can be found at:
https://keras.io/api/keras_cv/bounding_box/formats/
"""

import collections
from collections.abc import Sequence
import enum
from typing import Union

import numpy as np


# A BBox contains the bounding boox coordinates in the YXYX format, following
# this order: [top, left, bottom, right].
BBox = collections.namedtuple('BBox', 'ymin, xmin, ymax, xmax')


@enum.unique
class BBoxFormat(enum.Enum):
  """Supported bounding box formats by TFDS.

  If you add to this enum, you will also probably need to update the function
  `convert_coordinates_to_bbox`.
  """

  XYXY = 'XYXY'
  YXYX = 'YXYX'
  XYWH = 'XYWH'
  REL_XYXY = 'REL_XYXY'
  REL_YXYX = 'REL_YXYX'


BBoxFormatType = Union[BBoxFormat, str]


def convert_coordinates_to_bbox(
    coordinates: np.ndarray,
    input_format: BBoxFormatType = 'YXYX',
    img_shape=None,
    normalize: bool = True,
) -> BBox:
  """Converts four coordinates to a BBox using the YXYX format.

  Args:
    coordinates: np.array with the four coordinates expressing a bounding box.
    input_format: the format of the coordinates, e.g. 'XYXY' or 'XYWH'. See
      https://keras.io/api/keras_cv/bounding_box/formats/ for a list of popular
        bounding boxes formats.
    img_shape: 2-dimensions np.ndarray or iterable with (img_heigth, img_width).
    normalize: Whether to normalize the axes. If True, it will normalize the x
      coordinates using the image width, and the y coordinates using the image
      height. Defaults to True.

  Returns:
    A BBox in the correct format.

  Raises:
    ValueError if input_format is unknown, or if the
  """
  if len(coordinates) != 4:
    raise ValueError(f'Expected 4 coordinates, got {coordinates}.')
  coordinates = coordinates.astype(np.float64)

  try:
    if isinstance(input_format, str):
      input_format = BBoxFormat(input_format)
  except ValueError as e:
    raise ValueError(
        f'Unsupported bbox format: {format}. Currently supported bounding box'
        f' formats are: {[format.value for format in BBoxFormat]}'
    ) from e
  if (
      input_format == BBoxFormat.REL_XYXY or input_format == BBoxFormat.REL_YXYX
  ) and normalize:
    raise ValueError(
        'If the input format is normalized, then normalize should be False.'
    )
  if normalize and img_shape is None:
    raise ValueError(
        'If normalize is True, img_shape must be provided, but got None.'
    )
  if input_format == BBoxFormat.YXYX or input_format == BBoxFormat.REL_YXYX:
    bbox = BBox(
        ymin=coordinates[0],
        xmin=coordinates[1],
        ymax=coordinates[2],
        xmax=coordinates[3],
    )
  elif input_format == BBoxFormat.XYXY or input_format == BBoxFormat.REL_XYXY:
    bbox = BBox(
        ymin=coordinates[1],
        xmin=coordinates[0],
        ymax=coordinates[3],
        xmax=coordinates[2],
    )
  elif input_format == BBoxFormat.XYWH:
    x, y, width, height = coordinates
    bbox = BBox(ymin=y, xmin=x, ymax=(y + height), xmax=(x + width))
  else:
    raise ValueError(
        f'Unsupported bbox format: {format}. Currently supported bounding box'
        f' formats are: {[format.value for format in BBoxFormat]}'
    )

  if normalize:
    bbox = normalize_bbox(bbox, img_shape)

  return bbox


def normalize_bbox(
    bbox: BBox,
    img_shape: Union[np.ndarray, Sequence[float]],
) -> BBox:
  """Normalizes a bbox in the YXYX format.

  The x coordinates are normalized using the image's height, while the y
  coordinates using the image width.

  Args:
    bbox: The input BBox in YXYX format.
    img_shape: 2-dimensions np.ndarray or iterable with (imgage_heigth,
      image_width).

  Returns:
    Normalized BBox.

  Raises:
    ValueError if the img_shape shape or length is different than 2, or if any
    of the image shapes is zero.
  """
  if (isinstance(img_shape, np.ndarray) and img_shape.shape != (2,)) or (
      isinstance(img_shape, Sequence) and len(img_shape) != 2
  ):
    raise ValueError(f'Expected 2-dimensional img_shape, got {img_shape}.')

  height, width = img_shape[0], img_shape[1]
  if height == 0 or width == 0:
    raise ValueError(
        f'Valid image shapes cannot be zero, but got: {img_shape}.'
    )

  return BBox(
      ymin=bbox.ymin / height,
      xmin=bbox.xmin / width,
      ymax=bbox.ymax / height,
      xmax=bbox.xmax / width,
  )



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Meta register that uses other registers."""

from collections.abc import Iterable, Mapping, Sequence
import concurrent.futures
import difflib
import functools
import multiprocessing
import os
from typing import Any, Type

from absl import logging
from etils import epath
from tensorflow_datasets.core import dataset_builder
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import registered
from tensorflow_datasets.core.community import config as config_lib
from tensorflow_datasets.core.community import register_base
from tensorflow_datasets.core.community import register_package
from tensorflow_datasets.core.community import register_path
from tensorflow_datasets.core.utils import gcs_utils


def _load_registry(
    namespace: str,
    config: config_lib.NamespaceConfig,
) -> register_base.BaseRegister:
  """Returns a list of registers for the given paths."""
  code_paths = []
  data_paths = []
  for path in config.paths:
    if os.fspath(path).startswith('github'):
      code_paths.append(path)
    else:
      data_paths.append(path)

  if code_paths and data_paths:
    raise RuntimeError(
        f'Both a path containing code ({code_paths}) and a path containing data'
        f' ({data_paths} are specified. This is not supported'
    )

  if data_paths:
    return register_path.DataDirRegister(namespaces={namespace: config})
  elif code_paths:
    return register_package.PackageRegister(
        path=gcs_utils.GCS_COMMUNITY_INDEX_PATH
    )
  raise RuntimeError(f'No paths specified for {namespace}')


def _registers_per_namespace(
    config_per_namespace: Mapping[str, config_lib.NamespaceConfig],
) -> Mapping[str, Sequence[register_base.BaseRegister]]:
  return {
      namespace: [_load_registry(namespace, config)]
      for namespace, config in config_per_namespace.items()
  }


class DatasetRegistry(register_base.BaseRegister):
  """Registry of dataset registries.

  Each namespace is associated with one or more registers. Those registers can
  be any type of register, e.g. code on Github or datasets stored somewhere.

  Attributes:
    namespace_registry: registry of all the namespaces and their configs.
    config_per_namespace: mapping from namespace to configuration.
    registers_per_namespace: mapping from namespace to a sequence of registers
      for that namespace.
  """

  def __init__(self, namespace_registry: config_lib.NamespaceRegistry):
    self.namespace_registry = namespace_registry
    self._registers_per_namespace: dict[
        str, Sequence[register_base.BaseRegister]
    ] = {}
    self._is_registers_per_namespace_initialized = False

  @property
  def config_per_namespace(self) -> Mapping[str, config_lib.NamespaceConfig]:
    return self.namespace_registry.config_per_namespace

  def _init_registers_per_namespace(self):
    if self._is_registers_per_namespace_initialized:
      return
    self._registers_per_namespace.update(
        _registers_per_namespace(self.config_per_namespace)
    )
    self._is_registers_per_namespace_initialized = True

  @property
  def registers_per_namespace(
      self,
  ) -> Mapping[str, Sequence[register_base.BaseRegister]]:
    self._init_registers_per_namespace()
    return self._registers_per_namespace

  def add_namespace(
      self,
      namespace: str,
      config: config_lib.NamespaceConfig,
      registers: Sequence[register_base.BaseRegister],
  ):
    self._init_registers_per_namespace()
    if namespace in self._registers_per_namespace:
      raise ValueError(
          f'Namespace {namespace} already exists! Existing namespace config is'
          f' {self.config_per_namespace[namespace]}.'
      )
    self._registers_per_namespace[namespace] = registers
    self.namespace_registry.add_namespace(namespace=namespace, config=config)

  def has_namespace(self, namespace: str) -> bool:
    if not namespace:
      return False
    return namespace in self.config_per_namespace

  def list_namespaces(self) -> Sequence[str]:
    return sorted(self.config_per_namespace.keys())

  def list_builders(self) -> Sequence[str]:
    builders = []
    for registers in self.registers_per_namespace.values():
      for register in registers:
        builders.extend(register.list_builders())
    return builders

  def list_dataset_references(
      self, max_workers: int = multiprocessing.cpu_count()
  ) -> Iterable[naming.DatasetReference]:
    all_registers = []
    for registers in self.registers_per_namespace.values():
      all_registers.extend(registers)

    def _get_references(
        register: register_base.BaseRegister,
    ) -> Iterable[naming.DatasetReference]:
      try:
        yield from register.list_dataset_references()
      except Exception:  # pylint: disable=broad-except
        logging.exception(
            'Exception while getting dataset references from register %s',
            register,
        )

    with concurrent.futures.ThreadPoolExecutor(
        max_workers=max_workers
    ) as executor:
      output = list(executor.map(_get_references, all_registers))
      for dataset_references in output:
        yield from dataset_references

  def list_dataset_references_for_namespace(
      self, namespace: str
  ) -> Iterable[naming.DatasetReference]:
    """Lists the dataset references available for a specific namespace.

    Args:
      namespace: a namespace.

    Yields:
      An iterator over the datasets references for the given namespace.

    Raises:
      ValueError if the given namespace doesn't exist in the registry.
    """
    if not self.has_namespace(namespace):
      raise ValueError(f'Namespace {namespace} not found.')
    for register in self.registers_per_namespace[namespace]:
      try:
        yield from register.list_dataset_references()
      except Exception as e:  # pylint: disable=broad-except
        logging.exception(
            'Exception while getting dataset references from register %s: %s',
            register,
            e,
        )

  def list_builders_per_namespace(self, namespace: str) -> Sequence[str]:
    """Lists the builders available for a specific namespace."""
    builders = []
    if self.has_namespace(namespace):
      for register in self.registers_per_namespace[namespace]:
        builders.extend(register.list_builders())
    return builders

  def _get_list_builders_context(self, name: naming.DatasetName) -> str:
    """Adds relevant information to the error context."""
    # Add list of available datasets to error context.
    all_datasets = self.list_builders_per_namespace(name.namespace)
    all_datasets_str = '\n\t- '.join([''] + list(all_datasets))
    error_msg = (
        f'Available datasets under the same namespace:{all_datasets_str}\n'
    )
    # Add closest match to error context.
    close_matches = difflib.get_close_matches(str(name), all_datasets, n=1)
    if close_matches:
      error_msg += f'\nDid you mean: {name} -> {close_matches[0]} ?\n'
    return error_msg

  def _get_registers(
      self, name: naming.DatasetName
  ) -> Sequence[register_base.BaseRegister]:
    """Returns all available registers for a given namespace, if any.

    Args:
      name: str, the namespace's name.

    Raises:
      DatasetNotFound error if the namespace is not found.
    """
    if not self.has_namespace(name.namespace):
      error_msg = f'\nNamespace {name.namespace} not found. '
      error_msg += (
          'Note that the namespace should be one of: '
          f'{sorted(self.registers_per_namespace.keys())}.\n'
      )
      close_matches = difflib.get_close_matches(
          name.namespace, self.registers_per_namespace, n=1
      )
      if close_matches:
        error_msg += f'Did you mean: {name.namespace} -> {close_matches[0]} ?\n'
      raise registered.DatasetNotFoundError(error_msg)
    return self.registers_per_namespace[name.namespace]

  def builder_cls(
      self,
      name: naming.DatasetName,
  ) -> Type[dataset_builder.DatasetBuilder]:
    """Loads the builder class for the given dataset.

    Arguments:
      name: the name and namespace of the dataset to load the builder class for.

    Returns:
      DatasetNotFoundError if data is not found.
    """
    registers = self._get_registers(name)

    # Typically there's only 1, so add special case so that more informative
    # exceptions are raised.
    if len(registers) == 1:
      return registers[0].builder_cls(name)

    # If this dataset has multiple registers, use the first that can be found.
    for register in registers:
      try:
        return register.builder_cls(name)
      except registered.DatasetNotFoundError:
        pass

    raise registered.DatasetNotFoundError(
        f'Namespace {name.namespace} found, '
        f'but could not load dataset {name.name}.'
        f'{self._get_list_builders_context(name)}'
    )

  def builder(
      self,
      name: naming.DatasetName,
      **builder_kwargs: Any,
  ) -> dataset_builder.DatasetBuilder:
    """Loads the builder class for the given dataset."""
    registers = self._get_registers(name)

    # Typically there's only 1, so add special case so that more informative
    # exceptions are raised.
    if len(registers) == 1:
      return registers[0].builder(name, **builder_kwargs)

    if len(registers) > 1:
      raise ValueError(
          f'Namespace {name.namespace} has multiple registers! '
          f'This should not happen! Registers: {registers}'
      )

    raise registered.DatasetNotFoundError(
        f'Namespace {name.namespace} found with {len(registers)} registers, '
        f'but could not load dataset {name.name}.'
    )

  def get_builder_root_dirs(
      self, name: naming.DatasetName
  ) -> Sequence[epath.Path]:
    """Returns root dir of the generated builder (without version/config)."""
    result = []
    registers = self.registers_per_namespace[name.namespace]
    for register in registers:
      if isinstance(register, register_path.DataDirRegister):
        result.extend(register.get_builder_root_dirs(name))
      else:
        raise RuntimeError(f'Not supported for non datadir registers ({name})!')
    return result


@functools.lru_cache(maxsize=1)
def community_register() -> DatasetRegistry:
  return DatasetRegistry(config_lib.get_community_config())



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.features.image_feature."""

import contextlib
import enum
import os
import pathlib
from typing import Optional, Type

from absl.testing import parameterized
import numpy as np
import PIL
import pytest
import tensorflow as tf
import tensorflow_datasets as tfds
from tensorflow_datasets import testing
from tensorflow_datasets.core import features as features_lib

randint = np.random.randint


class LibWithImportError(enum.Enum):
  CV2 = 'cv2'
  PIL = 'pil'


@contextlib.contextmanager
def make_cv2_fail():
  """Recreates OpenCV not being installed during the test."""
  real_cv2 = tfds.features.image_feature.cv2
  tfds.features.image_feature.cv2 = None
  try:
    yield LibWithImportError.CV2
  finally:
    tfds.features.image_feature.cv2 = real_cv2


@contextlib.contextmanager
def make_pil_fail():
  """Recreates Pillow not being installed during the test."""
  real_pil = tfds.features.image_feature.PIL_Image
  tfds.features.image_feature.PIL_Image = None
  try:
    yield LibWithImportError.PIL
  finally:
    tfds.features.image_feature.PIL_Image = real_pil


@contextlib.contextmanager
def make_none_fail():
  """Used for a regular scenario, where both OpenCV and PIL are installed."""
  yield


def _unsupported_images_for_pil(
    np_dtype: Type[np.generic], failing_lib: Optional[LibWithImportError]
) -> bool:
  # PIL is used when OpenCV is not installed, but PIL doesn't support 16-bit
  # images.
  return np_dtype == np.uint16 and failing_lib == LibWithImportError.CV2


class ImageFeatureTest(
    testing.FeatureExpectationsTestCase, parameterized.TestCase
):

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail],
      dtypes=[
          (np.uint8, np.uint8),
          (np.uint16, np.uint16),
          (tf.uint8, np.uint8),
          (tf.uint16, np.uint16),
      ],
      channels=[1, 3, 4],
  )
  def test_images(self, make_lib_fail, dtypes, channels):
    dtype, np_dtype = dtypes
    with make_lib_fail() as failing_lib:
      if _unsupported_images_for_pil(np_dtype, failing_lib):
        return
      img = randint(256, size=(128, 100, channels), dtype=np_dtype)
      img_other_shape = randint(256, size=(64, 200, channels), dtype=np_dtype)

      filename = {
          np.uint8: {
              1: '6pixels_grayscale.png',
              3: '6pixels.png',
              4: '6pixels_4chan.png',
          },
          np.uint16: {
              1: '6pixels_grayscale_16bit.png',
              3: '6pixels_16bit.png',
              4: '6pixels_16bit_4chan.png',
          },
      }[np_dtype][channels]

      img_file_path = os.path.join(
          os.path.dirname(__file__), '../../testing/test_data', filename
      )
      with tf.io.gfile.GFile(img_file_path, 'rb') as f:
        img_byte_content = f.read()
      img_file_expected_content = np.array(
          [
              [[0, 255, 0, 255], [255, 0, 0, 255], [255, 0, 255, 255]],
              [[0, 0, 255, 255], [255, 255, 0, 255], [126, 127, 128, 255]],
          ],
          dtype=np_dtype,
      )[
          :, :, :channels
      ]  # Truncate (h, w, 4) -> (h, w, c)
      if dtype == np.uint16 or dtype == tf.uint16:
        img_file_expected_content *= 257  # Scale int16 images

      numpy_array = testing.FeatureExpectationItem(
          value=img,
          expected=img,
          expected_np=img,
      )
      file_path_as_string = testing.FeatureExpectationItem(
          value=img_file_path,
          expected=img_file_expected_content,
          expected_np=img_file_expected_content,
      )
      file_path_as_path = testing.FeatureExpectationItem(
          value=pathlib.Path(img_file_path),
          expected=img_file_expected_content,
          expected_np=img_file_expected_content,
      )
      images_bytes = testing.FeatureExpectationItem(
          value=img_byte_content,
          expected=img_file_expected_content,
          expected_np=img_file_expected_content,
      )
      img_shape_can_be_dynamic = testing.FeatureExpectationItem(
          value=img_other_shape,
          expected=img_other_shape,
          expected_np=img_other_shape,
      )
      invalid_type = testing.FeatureExpectationItem(
          value=randint(256, size=(128, 128, channels), dtype=np.uint32),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='dtype should be',
      )
      tests = [
          numpy_array,
          file_path_as_string,
          file_path_as_path,
          images_bytes,
          img_shape_can_be_dynamic,
          invalid_type,
      ]
      # PIL doesn't support 16-bit images.
      if (failing_lib != LibWithImportError.PIL) and np_dtype != np.uint16:
        tests.append(
            testing.FeatureExpectationItem(
                value=PIL.Image.open(img_file_path),
                expected=img_file_expected_content,
                expected_np=img_file_expected_content,
            )
        )
      self.assertFeature(
          feature=features_lib.Image(shape=(None, None, channels), dtype=dtype),
          shape=(None, None, channels),
          dtype=dtype,
          tests=tests,
          test_attributes=dict(
              _encoding_format=None,
              _use_colormap=False,
          ),
      )

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail],
      dtypes=[
          (np.uint8, np.uint8),
          (tf.uint8, np.uint8),
          (np.uint16, np.uint16),
          (tf.uint16, np.uint16),
      ],
      channels=[3, 4],
  )
  def test_images_with_invalid_shape(self, make_lib_fail, dtypes, channels):
    dtype, np_dtype = dtypes
    with make_lib_fail() as failing_lib:
      if _unsupported_images_for_pil(np_dtype, failing_lib):
        return
      invalid_number_of_dimensions = testing.FeatureExpectationItem(
          value=randint(256, size=(128, 128), dtype=np_dtype),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='must have the same rank',
      )
      invalid_number_of_channels = testing.FeatureExpectationItem(
          value=randint(256, size=(128, 128, 1), dtype=np_dtype),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='are incompatible',
      )
      self.assertFeature(
          feature=features_lib.Image(shape=(None, None, channels), dtype=dtype),
          shape=(None, None, channels),
          dtype=dtype,
          tests=[invalid_number_of_dimensions, invalid_number_of_channels],
          test_attributes=dict(
              _encoding_format=None,
              _use_colormap=False,
          ),
      )

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail]
  )
  def test_image_with_statically_defined_shape(self, make_lib_fail):
    with make_lib_fail():
      img_shaped = randint(256, size=(32, 64, 1), dtype=np.uint8)

      img_shape_should_be_conserved = testing.FeatureExpectationItem(
          value=img_shaped,
          expected=img_shaped,
          expected_np=img_shaped,
      )
      img_shape_should_be_static = testing.FeatureExpectationItem(
          value=randint(256, size=(31, 64, 1), dtype=np.uint8),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='are incompatible',
      )
      self.assertFeature(
          feature=features_lib.Image(
              shape=(32, 64, 1),
              encoding_format='png',
              use_colormap=True,
          ),
          shape=(32, 64, 1),
          dtype=np.uint8,
          tests=[img_shape_should_be_conserved, img_shape_should_be_static],
          test_attributes=dict(
              _encoding_format='png',
              _use_colormap=True,
          ),
      )

  @parameterized.product(
      make_lib_fail=[make_none_fail, make_cv2_fail, make_pil_fail],
      dtype=[np.float32, tf.float32],
  )
  def test_images_float(self, make_lib_fail, dtype):
    with make_lib_fail():
      img = np.random.rand(28, 28, 1).astype(np.float32)
      img_other_shape = np.random.rand(12, 34, 1).astype(np.float32)

      numpy_array = testing.FeatureExpectationItem(
          value=img,
          expected=img,
          expected_np=img,
      )
      img_shape_can_be_dynamic = testing.FeatureExpectationItem(
          value=img_other_shape,
          expected=img_other_shape,
          expected_np=img_other_shape,
      )
      invalid_type = testing.FeatureExpectationItem(
          value=img.astype(np.float64),
          raise_cls=ValueError,
          raise_cls_np=ValueError,
          raise_msg='dtype should be',
      )
      self.assertFeature(
          feature=features_lib.Image(shape=(None, None, 1), dtype=dtype),
          shape=(None, None, 1),
          dtype=dtype,
          tests=[
              numpy_array,
              img_shape_can_be_dynamic,
              invalid_type,
          ],
          test_attributes=dict(
              _encoding_format=None,
              _use_colormap=False,
          ),
      )

  @parameterized.product(
      dtypes=[
          (np.uint16, np.uint16),
          (tf.uint16, np.uint16),
      ],
      channels=[1, 3, 4],
  )
  def test_cannot_decode_16_bit_images_with_pil(self, dtypes, channels):
    dtype, np_dtype = dtypes
    # PIL is used when OpenCV fails.
    with make_cv2_fail():
      self.assertFeature(
          feature=features_lib.Image(shape=(128, 128, channels), dtype=dtype),
          shape=(128, 128, channels),
          dtype=dtype,
          tests=[
              testing.FeatureExpectationItem(
                  value=randint(256, size=(128, 128, channels), dtype=np_dtype),
                  raise_cls_np=ValueError,
                  raise_msg='PIL does not handle',
              )
          ],
      )

  def test_fail_if_neither_opencv_nor_pil_is_installed(self):
    with make_pil_fail():
      with make_cv2_fail():
        self.assertFeature(
            feature=features_lib.Image(shape=(128, 128, 3), dtype=np.uint8),
            shape=(128, 128, 3),
            dtype=np.uint8,
            tests=[
                testing.FeatureExpectationItem(
                    value=randint(256, size=(128, 128, 3), dtype=np.uint8),
                    raise_cls_np=ImportError,
                    raise_msg='requires either OpenCV or PIL',
                )
            ],
        )


@pytest.mark.parametrize(
    'shape, dtype, encoding_format, err_msg',
    [
        (None, np.uint16, r'jpeg', 'Acceptable `dtype` for jpeg:'),
        (None, tf.uint16, r'jpeg', 'Acceptable `dtype` for jpeg:'),
        (None, np.float32, None, 'only support single-channel'),
        (None, tf.float32, None, 'only support single-channel'),
        ((None, None, 1), np.float64, None, 'Acceptable `dtype`'),
        ((None, None, 1), tf.float64, None, 'Acceptable `dtype`'),
    ],
)
def test_invalid_img(shape, dtype, encoding_format, err_msg):
  with pytest.raises(ValueError, match=err_msg):
    features_lib.Image(
        shape=shape, dtype=dtype, encoding_format=encoding_format
    )


def test_pil_encoding():
  image_feat = features_lib.Image(encoding_format='png')
  x = np.uint8(np.array([[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]]))
  pil_image = PIL.Image.fromarray(x)
  pil_image.format = 'png'
  encoded_example = image_feat.encode_example(pil_image)

  assert isinstance(encoded_example, bytes)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ProteinNet dataset."""

from __future__ import annotations

import os
from typing import Dict, Iterator, List, Optional, Sequence, Tuple, Union
import urllib

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_Example = Dict[str, Union[int, str, np.ndarray, List]]
_ExampleIterator = Iterator[Tuple[str, _Example]]
_PROTEINNET_HOMEPAGE = 'https://github.com/aqlaboratory/proteinnet'

_LINES_PER_ENTRY = 33


def _parse_array(lines: Sequence[str]) -> np.ndarray:
  """Parse lines of tab-separated numbers into an array."""
  lines = [x.split('\t') for x in lines]
  return np.array(lines, dtype=np.float32)


def _parse_mask(line: str) -> np.ndarray:
  """Parse a string of `+` and `-` into a bool array."""
  return np.array([ch == '+' for ch in line], dtype=bool)


def _read_entry(fin: tf.io.gfile.GFile) -> Optional[Tuple[str, _Example]]:
  """Read an example from an input file.

  Args:
    fin: Input file object for reading dataset entries.

  Returns:
    The read exmple and its name, or None in cases of EOF.

  Raises:
    Exception: In case entry format is incorect.
  """
  lines = []
  for _ in range(_LINES_PER_ENTRY):
    line = fin.readline().strip()
    lines.append(line)

  if all(not line for line in lines):  # EOF?
    return None

  # Check structure.
  if (
      lines[0] != '[ID]'
      or lines[2] != '[PRIMARY]'
      or lines[4] != '[EVOLUTIONARY]'
      or lines[26] != '[TERTIARY]'
      or lines[30] != '[MASK]'
      or lines[32]
  ):
    raise ValueError('Incorrect data formatting.')
  lines = lines[:-1]  # Discard last empty (spacer) line.

  # The transposes below is required because TFDS allows unknown tensor
  # dimensions only in the first axis.
  key = lines[1]
  example = {
      'id': key,
      'primary': list(lines[3]),
      'evolutionary': _parse_array(lines[5:26]).transpose(),
      'tertiary': _parse_array(lines[27:30]).transpose(),
      'mask': _parse_mask(lines[31]),
      'length': len(lines[3]),
  }
  return key, example


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for the ProteinNet dataset."""

  URL = (
      'https://sharehost.hms.harvard.edu/sysbio/alquraishi/proteinnet'
      '/human_readable/'
  )
  FILES = {
      'casp7': 'casp7.tar.gz',
      'casp8': 'casp8.tar.gz',
      'casp9': 'casp9.tar.gz',
      'casp10': 'casp10.tar.gz',
      'casp11': 'casp11.tar.gz',
      'casp12': 'casp12.tar.gz',
  }
  THRESHOLDS = [30, 50, 70, 90, 95, 100]
  AMINOACIDS = [
      'A',
      'C',
      'D',
      'E',
      'F',
      'G',
      'H',
      'I',
      'K',
      'L',
      'M',
      'N',
      'P',
      'Q',
      'R',
      'S',
      'T',
      'V',
      'W',
      'Y',
  ]

  BUILDER_CONFIGS = [
      tfds.core.BuilderConfig(name='casp7'),
      tfds.core.BuilderConfig(name='casp8'),
      tfds.core.BuilderConfig(name='casp9'),
      tfds.core.BuilderConfig(name='casp10'),
      tfds.core.BuilderConfig(name='casp11'),
      tfds.core.BuilderConfig(name='casp12'),
  ]

  VERSION = tfds.core.Version('1.0.0')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            'id': tfds.features.Text(),
            'primary': tfds.features.Sequence(
                tfds.features.ClassLabel(names=self.AMINOACIDS)
            ),
            'evolutionary': tfds.features.Tensor(
                shape=(None, 21), dtype=np.float32
            ),
            'tertiary': tfds.features.Tensor(shape=(None, 3), dtype=np.float32),
            'mask': tfds.features.Tensor(shape=(None,), dtype=np.bool_),
            'length': tfds.features.Tensor(shape=(), dtype=np.int32),
        }),
        supervised_keys=('primary', 'tertiary'),
        homepage=_PROTEINNET_HOMEPAGE,
    )

  def _split_generators(
      self, dl_manager: tfds.download.DownloadManager
  ) -> Dict[Union[str, tfds.Split], _ExampleIterator]:
    """Returns SplitGenerators."""
    name = self.builder_config.name  # Configurable dataset (config) name.
    path = dl_manager.download_and_extract(
        urllib.parse.urljoin(self.URL, self.FILES[name])
    )

    splits = {
        tfds.Split.VALIDATION: self._generate_examples(
            os.path.join(path, name, 'validation')
        ),
        tfds.Split.TEST: self._generate_examples(
            os.path.join(path, name, 'testing')
        ),
    }
    for threshold in self.THRESHOLDS:  # Train splits.
      split_path = os.path.join(path, name, f'training_{threshold}')
      splits[f'train_{threshold}'] = self._generate_examples(split_path)
    return splits

  def _generate_examples(self, filename: str) -> _ExampleIterator:
    """Yields examples."""
    with tf.io.gfile.GFile(filename, mode='r') as fin:
      while True:
        example = _read_entry(fin)
        if example is None:
          break
        yield example



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cityscapes Datasets."""

import os
import re

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds


class CityscapesConfig(tfds.core.BuilderConfig):
  """BuilderConfig for Cityscapes.

  Args:
    right_images (bool): Enables right images for stereo image tasks.
    segmentation_labels (bool): Enables image segmentation labels.
    disparity_maps (bool): Enables disparity maps.
    train_extra_split (bool): Enables train_extra split. This automatically
      enables coarse grain segmentations, if segmentation labels are used.
  """

  def __init__(
      self,
      *,
      right_images=False,
      segmentation_labels=True,
      disparity_maps=False,
      train_extra_split=False,
      **kwargs,
  ):
    super(CityscapesConfig, self).__init__(version='1.0.0', **kwargs)

    self.right_images = right_images
    self.segmentation_labels = segmentation_labels
    self.disparity_maps = disparity_maps
    self.train_extra_split = train_extra_split

    self.ignored_ids = set()

    # Setup required zips and their root dir names
    self.zip_root = {}
    self.zip_root['images_left'] = (
        'leftImg8bit_trainvaltest.zip',
        'leftImg8bit',
    )

    if self.train_extra_split:
      self.zip_root['images_left/extra'] = (
          'leftImg8bit_trainextra.zip',
          'leftImg8bit',
      )

    if self.right_images:
      self.zip_root['images_right'] = (
          'rightImg8bit_trainvaltest.zip',
          'rightImg8bit',
      )
      if self.train_extra_split:
        self.zip_root['images_right/extra'] = (
            'rightImg8bit_trainextra.zip',
            'rightImg8bit',
        )

    if self.segmentation_labels:
      if not self.train_extra_split:
        self.zip_root['segmentation_labels'] = (
            'gtFine_trainvaltest.zip',
            'gtFine',
        )
        self.label_suffix = 'gtFine_labelIds'
      else:
        # The 'train extra' split only has coarse labels unlike train and val.
        # Therefore, for consistency across splits, we also enable coarse labels
        # using the train_extra_split flag.
        self.zip_root['segmentation_labels'] = ('gtCoarse.zip', 'gtCoarse')
        self.zip_root['segmentation_labels/extra'] = (
            'gtCoarse.zip',
            'gtCoarse',
        )
        self.label_suffix = 'gtCoarse_labelIds'

    if self.disparity_maps:
      self.zip_root['disparity_maps'] = (
          'disparity_trainvaltest.zip',
          'disparity',
      )
      if self.train_extra_split:
        self.zip_root['disparity_maps/extra'] = (
            'disparity_trainextra.zip',
            'disparity',
        )
        # No disparity for this file
        self.ignored_ids.add('troisdorf_000000_000073')


class Builder(tfds.core.GeneratorBasedBuilder):
  """Base class for Cityscapes datasets."""

  MANUAL_DOWNLOAD_INSTRUCTIONS = """\
  You have to download files from https://www.cityscapes-dataset.com/login/
  (This dataset requires registration).
  For basic config (semantic_segmentation) you must download
  'leftImg8bit_trainvaltest.zip' and 'gtFine_trainvaltest.zip'.
  Other configs do require additional files - please see code for more details.
  """

  BUILDER_CONFIGS = [
      CityscapesConfig(
          name='semantic_segmentation',
          description='Cityscapes semantic segmentation dataset.',
          right_images=False,
          segmentation_labels=True,
          disparity_maps=False,
          train_extra_split=False,
      ),
      CityscapesConfig(
          name='semantic_segmentation_extra',
          description=(  # pylint: disable=line-too-long
              'Cityscapes semantic segmentation dataset with train_extra split'
              ' and coarse labels.'
          ),
          right_images=False,
          segmentation_labels=True,
          disparity_maps=False,
          train_extra_split=True,
      ),
      CityscapesConfig(
          name='stereo_disparity',
          description='Cityscapes stereo image and disparity maps dataset.',
          right_images=True,
          segmentation_labels=False,
          disparity_maps=True,
          train_extra_split=False,
      ),
      CityscapesConfig(
          name='stereo_disparity_extra',
          description=(  # pylint: disable=line-too-long
              'Cityscapes stereo image and disparity maps dataset with'
              ' train_extra split.'
          ),
          right_images=True,
          segmentation_labels=False,
          disparity_maps=True,
          train_extra_split=True,
      ),
  ]

  def _info(self):
    # Enable features as necessary
    features = {}
    features['image_id'] = tfds.features.Text()
    features['image_left'] = tfds.features.Image(
        shape=(1024, 2048, 3), encoding_format='png'
    )

    if self.builder_config.right_images:
      features['image_right'] = tfds.features.Image(
          shape=(1024, 2048, 3), encoding_format='png'
      )

    if self.builder_config.segmentation_labels:
      features['segmentation_label'] = tfds.features.Image(
          shape=(1024, 2048, 1), encoding_format='png', use_colormap=True
      )

    if self.builder_config.disparity_maps:
      features['disparity_map'] = tfds.features.Image(
          shape=(1024, 2048, 1), encoding_format='png'
      )

    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(features),
        homepage='https://www.cityscapes-dataset.com',
    )

  def _split_generators(self, dl_manager):
    paths = {}
    for split, (zip_file, _) in self.builder_config.zip_root.items():
      paths[split] = os.path.join(dl_manager.manual_dir, zip_file)

    if any(not tf.io.gfile.exists(z) for z in paths.values()):
      msg = 'You must download the dataset files manually and place them in: '
      msg += ', '.join(paths.values())
      raise AssertionError(msg)

    for split, (_, zip_root) in self.builder_config.zip_root.items():
      paths[split] = os.path.join(dl_manager.extract(paths[split]), zip_root)

    splits = [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                feat_dir: os.path.join(path, 'train')
                for feat_dir, path in paths.items()
                if not feat_dir.endswith('/extra')
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                feat_dir: os.path.join(path, 'val')
                for feat_dir, path in paths.items()
                if not feat_dir.endswith('/extra')
            },
        ),
    ]

    # Test split does not exist in coarse dataset
    if not self.builder_config.train_extra_split:
      splits.append(
          tfds.core.SplitGenerator(
              name=tfds.Split.TEST,
              gen_kwargs={
                  feat_dir: os.path.join(path, 'test')
                  for feat_dir, path in paths.items()
                  if not feat_dir.endswith('/extra')
              },
          )
      )
    else:
      splits.append(
          tfds.core.SplitGenerator(
              name='train_extra',
              gen_kwargs={  # pylint: disable=g-complex-comprehension
                  feat_dir.replace('/extra', ''): os.path.join(
                      path, 'train_extra'
                  )
                  for feat_dir, path in paths.items()
                  if feat_dir.endswith('/extra')
              },
          )
      )
    return splits

  def _generate_examples(self, **paths):
    left_imgs_root = paths['images_left']
    for city_id in tf.io.gfile.listdir(left_imgs_root):
      paths_city_root = {
          feat_dir: os.path.join(path, city_id)
          for feat_dir, path in paths.items()
      }

      left_city_root = paths_city_root['images_left']
      for left_img in tf.io.gfile.listdir(left_city_root):
        left_img_path = os.path.join(left_city_root, left_img)
        image_id = _get_left_image_id(left_img)

        if image_id in self.builder_config.ignored_ids:
          continue

        features = {
            'image_id': image_id,
            'image_left': left_img_path,
        }

        if self.builder_config.right_images:
          features['image_right'] = os.path.join(
              paths_city_root['images_right'],
              '{}_rightImg8bit.png'.format(image_id),
          )

        if self.builder_config.segmentation_labels:
          features['segmentation_label'] = os.path.join(
              paths_city_root['segmentation_labels'],
              '{}_{}.png'.format(image_id, self.builder_config.label_suffix),
          )

        if self.builder_config.disparity_maps:
          features['disparity_map'] = os.path.join(
              paths_city_root['disparity_maps'],
              '{}_disparity.png'.format(image_id),
          )

        yield image_id, features


# Helper functions

LEFT_IMAGE_FILE_RE = re.compile(r'([a-z\-]+)_(\d+)_(\d+)_leftImg8bit\.png')


def _get_left_image_id(left_image):
  """Returns the id of an image file.

  Used to associate an image file with its corresponding label.
  Example:
    'bonn_000001_000019_leftImg8bit' -> 'bonn_000001_000019'

  Args:
    left_image: name of the image file.

  Returns:
    Id of the image (see example above).
  """
  match = LEFT_IMAGE_FILE_RE.match(left_image)
  return '{}_{}_{}'.format(*match.groups())



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Async download API with checksum verification. No business logic."""

from __future__ import annotations

from collections.abc import Iterable, Iterator
import concurrent.futures
import contextlib
import dataclasses
import functools
import hashlib
import io
import os
import re
import typing
from typing import Any, ContextManager
import urllib

from etils import epath
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import units
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.download import checksums as checksums_lib
from tensorflow_datasets.core.download import resource as resource_lib
from tensorflow_datasets.core.download import util as download_utils_lib
from tensorflow_datasets.core.utils import tqdm_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import promise
from tensorflow_datasets.core.utils.lazy_imports_utils import requests
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


_DRIVE_URL = re.compile(r'^https://drive\.google\.com/')
MAX_RETRIES = 10

if typing.TYPE_CHECKING:
  # Response interface. Has `.url` and `.headers` attribute
  Response = requests.Response | urllib.response.addinfourl
else:
  Response = Any


@dataclasses.dataclass(eq=False, frozen=True)
class DownloadResult:
  path: epath.Path
  url_info: checksums_lib.UrlInfo


@utils.memoize()
def get_downloader(*args: Any, **kwargs: Any) -> _Downloader:
  return _Downloader(*args, **kwargs)


def read_url_info(url_path: epath.Path) -> checksums_lib.UrlInfo:
  """Loads the `UrlInfo` from the `.INFO` file."""
  file_info = resource_lib.read_info_file(url_path)
  if 'url_info' not in file_info:
    raise ValueError(
        'Could not find `url_info` in {}. This likely indicates that '
        'the files where downloaded with a previous version of TFDS (<=3.1.0). '
    )
  url_info = file_info['url_info']
  url_info.setdefault('filename', None)
  url_info['size'] = utils.Size(url_info['size'])
  return checksums_lib.UrlInfo(**url_info)


def _filename_from_content_disposition(
    content_disposition: str,
) -> str | None:
  """Extract the filename from the content disposition.

  Parse the content_definition as defined in:
  https://tools.ietf.org/html/rfc2616

  Note:

   * If both encoded (`filename*=`) and ascii (filename=) name are defined,
     the function returns the ascii name, as encoding might create issue on
     some systems
   * If only the encoded name is defined (e.g.
     `filename*=UTF-8''%e2%82%ac.txt`), the function return None as this is
     not yet supported.

  Args:
      content_disposition: String to parse.

  Returns:
      filename: The filename, or None if filename could not be parsed
  """
  match = re.findall(
      # Regex (see unittests for examples):
      # ` *` : Strip eventual whitespaces
      # `['"]?` : Filename is optionally wrapped in quote
      # `([^;\r\n"']+)` : Filename can be any symbol except those
      # `;?` : Stop when encountering optional `;`
      r"""filename= *['"]?([^;\r\n"']+)['"]? *;?""",
      content_disposition,
      flags=re.IGNORECASE,
  )
  if not match:
    return None
  elif len(match) != 1:
    raise ValueError(
        f'Error while parsing filename for: {content_disposition}\n'
        f'Multiple filename detected: {list(match)}'
    )
  return os.path.basename(match[0].rstrip())


def _get_filename(response: Response) -> str:
  content_disposition = response.headers.get('content-disposition', None)
  if content_disposition:
    filename = _filename_from_content_disposition(content_disposition)
    if filename:
      return filename
  # Otherwise, fallback on extracting the name from the url.
  return _basename_from_url(response.url)


def _process_gdrive_confirmation(original_url: str, contents: str) -> str:
  """Process Google Drive confirmation page.

  Extracts the download link from a Google Drive confirmation page.

  Args:
      original_url: The URL the confirmation page was originally retrieved from.
      contents: The confirmation page's HTML.

  Returns:
      download_url: The URL for downloading the file.
  """
  bs4 = lazy_imports_lib.lazy_imports.bs4
  soup = bs4.BeautifulSoup(contents, 'html.parser')
  form = soup.find('form')
  if not form:
    raise ValueError(
        f'Failed to obtain confirmation link for GDrive URL {original_url}.'
    )
  action = form.get('action', '')
  if not action:
    raise ValueError(
        f'Failed to obtain confirmation link for GDrive URL {original_url}.'
    )
  # Find the <input>s named 'uuid', 'export', 'id' and 'confirm'
  input_names = ['uuid', 'export', 'id', 'confirm']
  params = {}
  for name in input_names:
    input_tag = form.find('input', {'name': name})
    if input_tag:
      params[name] = input_tag.get('value', '')
  query_string = urllib.parse.urlencode(params)
  download_url = f'{action}?{query_string}' if query_string else action
  download_url = urllib.parse.urljoin(original_url, download_url)
  return download_url


class _Downloader:
  """Class providing async download API with checksum validation.

  Do not instantiate this class directly. Instead, call `get_downloader()`.
  """

  _DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS = 50
  _pbar_url: tqdm_utils._TqdmPbarAsync
  _pbar_dl_size: tqdm_utils._TqdmPbarAsync

  def __init__(
      self,
      max_simultaneous_downloads: (
          int | None
      ) = _DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS,
      checksumer=None,
  ):
    """Init _Downloader instance.

    Args:
      max_simultaneous_downloads: Optional max number of simultaneous downloads.
        If None then it defaults to `self._DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS`.
      checksumer: `hashlib.HASH`. Defaults to `hashlib.sha256`.
    """
    self._executor = concurrent.futures.ThreadPoolExecutor(
        max_workers=max_simultaneous_downloads
        or self._DEFAULT_MAX_SIMULTANEOUS_DOWNLOADS
    )
    self._checksumer_cls = checksumer or hashlib.sha256

  @contextlib.contextmanager
  def tqdm(self) -> Iterator[None]:
    """Add a progression bar for the current download."""
    async_tqdm = utils.async_tqdm
    with async_tqdm(
        total=0, desc='Dl Completed...', unit=' url', mininterval=1.0
    ) as pbar_url:
      with async_tqdm(
          total=0, desc='Dl Size...', unit=' MiB', mininterval=1.0
      ) as pbar_dl_size:
        self._pbar_url = pbar_url
        self._pbar_dl_size = pbar_dl_size
        yield

  def increase_tqdm(self, url_info: checksums_lib.UrlInfo) -> None:
    """Update the tqdm bars to visually indicate the url_info is downloaded."""
    self._pbar_url.update_total(1)
    self._pbar_url.update(1)
    self._pbar_dl_size.update_total(url_info.size)
    self._pbar_dl_size.update(url_info.size)

  def download(
      self, url: str, destination_path: epath.Path, verify: bool = True
  ) -> promise.Promise[concurrent.futures.Future[DownloadResult]]:
    """Download url to given path.

    Returns Promise -> sha256 of downloaded file.

    Args:
      url: Address of resource to download.
      destination_path: Path to directory where to download the resource.
      verify: Whether to verify ssl certificates

    Returns:
      Promise obj -> Download result.
    """
    self._pbar_url.update_total(1)
    future = self._executor.submit(
        self._sync_download, url, destination_path, verify
    )
    return promise.Promise.resolve(future)

  def _sync_file_copy(
      self,
      filepath: str,
      destination_path: epath.Path,
  ) -> DownloadResult:
    """Downloads the file through `tf.io.gfile` API."""
    filename = os.path.basename(filepath)
    out_path = destination_path / filename
    tf.io.gfile.copy(filepath, out_path)
    url_info = checksums_lib.compute_url_info(
        out_path, checksum_cls=self._checksumer_cls
    )
    self._pbar_dl_size.update_total(url_info.size)
    self._pbar_dl_size.update(url_info.size)
    self._pbar_url.update(1)
    return DownloadResult(path=out_path, url_info=url_info)

  def _sync_download(
      self, url: str, destination_path: epath.Path, verify: bool = True
  ) -> DownloadResult:
    """Synchronous version of `download` method.

    To download through a proxy, the `HTTP_PROXY`, `HTTPS_PROXY`,
    `REQUESTS_CA_BUNDLE`,... environment variables can be exported, as
    described in:
    https://requests.readthedocs.io/en/master/user/advanced/#proxies

    Args:
      url: Url to download.
      destination_path: Path where to write it.
      verify: Whether to verify ssl certificates.

    Returns:
      Download result.

    Raises:
      DownloadError: when download fails.
    """
    try:
      # If url is on a filesystem that gfile understands, use copy. Otherwise,
      # use requests (http) or urllib (ftp).
      if not url.startswith('http'):
        return self._sync_file_copy(url, destination_path)
    except tf.errors.UnimplementedError:
      pass

    with _open_url(url, verify=verify) as (response, iter_content):
      fname = _get_filename(response)
      path = destination_path / fname
      size = 0

      # Initialize the download size progress bar
      size_mb = 0
      unit_mb = units.MiB
      total_size = int(response.headers.get('Content-length', 0)) // unit_mb
      self._pbar_dl_size.update_total(total_size)
      with path.open('wb') as file_:
        checksum = self._checksumer_cls()
        for block in iter_content:
          size += len(block)
          checksum.update(block)
          file_.write(block)

          # Update the download size progress bar
          size_mb += len(block)
          if size_mb > unit_mb:
            self._pbar_dl_size.update(size_mb // unit_mb)
            size_mb %= unit_mb
    self._pbar_url.update(1)
    return DownloadResult(
        path=path,
        url_info=checksums_lib.UrlInfo(
            checksum=checksum.hexdigest(),
            size=utils.Size(size),
            filename=fname,
        ),
    )


def _open_url(
    url: str,
    **kwargs: Any,
) -> ContextManager[tuple[Response, Iterable[bytes]]]:
  """Context manager to open an url.

  Args:
    url: The url to open
    **kwargs: Additional kwargs to forward to `request.get`.

  Returns:
    response: The url response with `.url` and `.header` attributes.
    iter_content: A `bytes` iterator which yield the content.
  """
  # Download FTP urls with `urllib`, otherwise use `requests`
  open_fn = _open_with_urllib if url.startswith('ftp') else _open_with_requests
  return open_fn(url, **kwargs)


@contextlib.contextmanager
def _open_with_requests(
    url: str,
    **kwargs: Any,
) -> Iterator[tuple[Response, Iterable[bytes]]]:
  """Open url with request."""
  with requests.Session() as session:
    retries = requests.packages.urllib3.util.retry.Retry(
        total=MAX_RETRIES,
        backoff_factor=0.2,
        status_forcelist=[500, 502, 503, 504],
        raise_on_redirect=True,
        raise_on_status=True,
    )
    session.mount('http://', requests.adapters.HTTPAdapter(max_retries=retries))
    session.mount(
        'https://', requests.adapters.HTTPAdapter(max_retries=retries)
    )
    with session.get(url, stream=True, **kwargs) as response:
      if (
          _DRIVE_URL.match(url)
          and 'Content-Disposition' not in response.headers
      ):
        download_url = _process_gdrive_confirmation(url, response.text)
        with session.get(
            download_url, stream=True, **kwargs
        ) as download_response:
          _assert_status(download_response)
          yield (
              download_response,
              download_response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE),
          )
      else:
        _assert_status(response)
        yield (
            response,
            response.iter_content(chunk_size=io.DEFAULT_BUFFER_SIZE),
        )


@contextlib.contextmanager
def _open_with_urllib(
    url: str,
    **kwargs: Any,
) -> Iterator[tuple[Response, Iterable[bytes]]]:
  del kwargs
  with urllib.request.urlopen(url) as response:  # pytype: disable=attribute-error
    yield (
        response,
        iter(functools.partial(response.read, io.DEFAULT_BUFFER_SIZE), b''),
    )


def _assert_status(response: requests.Response) -> None:
  """Ensure the URL response is 200."""
  if response.status_code != 200:
    raise download_utils_lib.DownloadError(
        'Failed to get url {}. HTTP code: {}.'.format(
            response.url, response.status_code
        )
    )


def _basename_from_url(url: str) -> str:
  """Returns file name of file at given url."""
  filename = urllib.parse.urlparse(url).path
  filename = os.path.basename(filename)
  # Replace `%2F` (html code for `/`) by `_`.
  # This is consistent with how Chrome rename downloaded files.
  filename = filename.replace('%2F', '_')
  return filename or 'unknown_name'



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NumPy utils."""

from typing import Any, Callable, Tuple, Union

import numpy as np
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tree


@py_utils.memoize()
def to_np_shape(shape: type_utils.Shape, replace_int=-1) -> Tuple[int]:
  return tuple(dim if dim is not None else replace_int for dim in shape)


def _dict_to_list(dict_of_lists: dict[str, Any]) -> list[Any]:
  """Converts a dict of lists to a list of dicts with the same structure."""
  feature_name = next(iter(dict_of_lists.keys()))
  length = len(dict_of_lists[feature_name])
  return [
      {k: dict_of_lists[k][i] for k in dict_of_lists} for i in range(length)
  ]


def _list_to_dict(
    list_of_elements: list[Any],
) -> Union[dict[str, Any], np.ndarray]:
  """Converts a list of elements to a dict of elements with the same struct."""
  if not list_of_elements:
    return {}
  first = list_of_elements[0]
  if isinstance(first, dict):
    return {k: np.asarray([v[k] for v in list_of_elements]) for k in first}
  else:
    return np.asarray(list_of_elements)


def np_map_fn(
    fn: Callable[[Any], Any],
    slices: Union[list[Any], dict[str, Any], np.ndarray],
):
  """Executes fn on all slices. NumPy equivalent of tf.map_fn.

  Args:
    fn: The function to execute.
    slices: An iterable or a dict of iterables.

  Returns:
    An object with the same structure as `slices` where every slice has `fn`
      applied.
  """
  if isinstance(slices, (list, np.ndarray)):
    return [fn(example) for example in slices]
  elif isinstance(slices, dict):
    if not slices:
      return slices
    if any(tree.is_nested(value) for value in slices.values()):
      raise NotImplementedError("Nested dict not supported by np_map_fn.")
    # For each slice of the dictionary...
    # {             <----- slice1 ----->     <----- slice2 ----->
    #     "image": [b"\x89PNG\r\n\x1a\n...", b"\x89PNG\r\n\x1a\n...", ...]
    #     "shape": [(100, 100, 3),           (100, 100, 3),           ...]
    # }
    # ...we apply the `fn` and reconstruct the dictionary:
    # {
    #     "image": np.array([[[255, 255, 0], ...]], ...])
    #     "shape": [(100, 100, 3),                  ...]
    # }
    fn_slices = [fn(s) for s in _dict_to_list(slices)]
    return _list_to_dict(fn_slices)
  else:
    raise ValueError(f"type {type(slices)} is not supported by np_map_fn.")



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""real_toxicity_prompts dataset."""
import json
from etils import epath
import numpy as np

import tensorflow_datasets.public_api as tfds


# Features describing toxicity annotations of prompts and continuations.
_TOXIC_FEATURES = tfds.features.FeaturesDict({
    "text": tfds.features.Text(),
    "profanity": np.float32,
    "sexually_explicit": np.float32,
    "identity_attack": np.float32,
    "flirtation": np.float32,
    "threat": np.float32,
    "insult": np.float32,
    "severe_toxicity": np.float32,
    "toxicity": np.float32,
})


class Builder(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for real_toxicity_prompts dataset."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""

    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict({
            "filename": tfds.features.Text(),
            "begin": np.int32,
            "end": np.int32,
            "challenging": np.bool_,
            "prompt": _TOXIC_FEATURES,
            "continuation": _TOXIC_FEATURES,
        }),
        supervised_keys=None,
        homepage="https://github.com/allenai/real-toxicity-prompts",
        license="https://github.com/allenai/real-toxicity-prompts/blob/master/LICENSE",
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(
        "https://ai2-public-datasets.s3.amazonaws.com/realtoxicityprompts/realtoxicityprompts-data.tar.gz"
    )

    return {
        "train": self._generate_examples(
            path / "realtoxicityprompts-data/prompts.jsonl"
        ),
    }

  def _generate_examples(self, path):
    """Yields examples."""

    def _get_toxic_features(row_toxic_data):
      return {f: row_toxic_data[f] for f in _TOXIC_FEATURES.keys()}

    def _get_row_features(row_data):
      row_features = {
          "filename": row_data["filename"],
          "begin": row_data["begin"],
          "end": row_data["end"],
          "challenging": row_data["challenging"],
          "prompt": _get_toxic_features(row_data["prompt"]),
          "continuation": _get_toxic_features(row_data["continuation"]),
      }
      return row_features

    with epath.Path(path).open() as f:
      for i, row in enumerate(f):
        row_data = json.loads(row)
        yield i, _get_row_features(row_data)



# coding=utf-8
# Copyright 2025 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BinarizedMNIST."""

import numpy as np
from six.moves import urllib
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.image_classification import mnist
import tensorflow_datasets.public_api as tfds

_URL = "http://www.cs.toronto.edu/~larocheh/public/datasets/binarized_mnist/"
_TRAIN_DATA_FILENAME = "binarized_mnist_train.amat"
_VALID_DATA_FILENAME = "binarized_mnist_valid.amat"
_TEST_DATA_FILENAME = "binarized_mnist_test.amat"


class Builder(tfds.core.GeneratorBasedBuilder):
  """A specific binarization of the MNIST dataset."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial Release",
  }

  def _info(self):
    return self.dataset_info_from_configs(
        features=tfds.features.FeaturesDict(
            {"image": tfds.features.Image(shape=mnist.MNIST_IMAGE_SHAPE)}
        ),
        homepage="http://www.dmi.usherb.ca/~larocheh/mlpython/_modules/datasets/binarized_mnist.html",
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    filenames = {
        "train_data": _TRAIN_DATA_FILENAME,
        "validation_data": _VALID_DATA_FILENAME,
        "test_data": _TEST_DATA_FILENAME,
    }
    files = dl_manager.download(
        {k: urllib.parse.urljoin(_URL, v) for k, v in filenames.items()}
    )

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs=dict(
                data_path=files["train_data"],
            ),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs=dict(
                data_path=files["validation_data"],
            ),
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs=dict(
                data_path=files["test_data"],
            ),
        ),
    ]

  def _generate_examples(self, data_path):
    """Generate Binarized MNIST examples as dicts.

    Args:
      data_path (str): Path to the data files

    Yields:
      Generator yielding the next examples
    """
    with tf.io.gfile.GFile(data_path, "rb") as f:
      images = (
          np.loadtxt(f, delimiter=" ")
          .astype(np.uint8)
          .reshape((-1,) + mnist.MNIST_IMAGE_SHAPE)
      )
    for index, image in enumerate(images):
      yield index, {"image": image}
